
from pathlib import Path as _Path

__all__ = ["OutputFiles"]


def _get_bool(arg):
    """Simple function to make sure that flags that are supposed
       to be true/false are actually stored as bools
    """
    if arg:
        return True
    else:
        return False


def _is_empty(outdir):
    """Simple function that checks whether or not the passed directory
       is empty
    """
    import os

    if os.path.isfile(outdir):
        # this is a file, so it definitely is not 'empty'
        return False
    elif os.path.isdir(outdir):
        # the directory is empty if it contains no hidden files
        # or other directories - sorry to wipe out dotfiles...!
        return len(os.listdir(outdir)) == 0
    else:
        return True


def _rmdir(directory):
    """Function modified from one copied from 'mitch' on stackoverflow
       https://stackoverflow.com/questions/13118029/deleting-folders-in-python-recursively
    """
    directory = _Path(directory)

    # first, check for removing important directories such as $HOME or root
    if directory == _Path.home():
        raise FileExistsError(f"We WILL NOT remove your "
                              f"home directory ${directory}")

    if directory == _Path("/"):
        raise FileExistsError(f"We WILL NOT remove the root directory "
                              f"{directory}")

    # get the directory containing '$HOME'
    if directory == _Path.home().parent:
        raise FileExistsError(f"We WILL NOT remove the users/home "
                              f"directory {directory}")

    if not directory.is_dir():
        directory.unlink()
        return

    from .utils._console import Console

    for item in directory.iterdir():
        if item.is_dir():
            _rmdir(item)
        else:
            item.unlink()

    Console.print(f"removing directory {directory}", style="warning")
    directory.rmdir()


def _check_remove(outdir, prompt):
    """Function to check if the user wants to remove the directory,
       giving them the option to continue, quit or remove all files
    """
    if prompt is None:
        raise FileExistsError(f"Cannot continue as {outdir} already exists!")

    from .utils._console import Console
    Console.warning(f"{outdir} already exists.")
    y = prompt("Do you want to remove it? (y/n) ")

    y = y.strip().lower()

    if len(y) > 0 and y == "y":
        Console.print(f"Removing all files in {outdir}", style="warning")
        _rmdir(_Path(outdir))
        return

    Console.warning(f"Continuing with this run will mix its output with "
                    f"the files already in {outdir}.")

    y = prompt("Do you want to continue with this run? (y/n) ")

    y = y.strip().lower()

    if len(y) == 0 or y != "y":
        from .utils._console import Console
        Console.error(f"Exiting the program as we cannot run any more.")
        import sys
        sys.exit(-1)


def _force_remove(outdir, prompt):
    """Function to force the removal of a directory, using the
       passed prompt to double-check with the user. If 'prompt'
       is None, then we go ahead
    """
    import os
    if not os.path.exists(outdir):
        return

    from .utils._console import Console

    if prompt:
        Console.warning(f"{outdir} already exists")

        y = prompt("Do you want to remove it? (y/n) ")
        y = y.strip().lower()

        if len(y) == 0 or y != "y":
            raise FileExistsError(
                f"Cannot continue as {outdir} already exists")

    Console.print(f"Removing all files in {outdir}", style="red")
    _rmdir(_Path(outdir))


def _expand(path):
    """Expand all variables and user indicators in the passed path"""
    import os
    return os.path.expanduser(os.path.expandvars(path))


class OutputFiles:
    """This is a class that manages all of the output files that
       are written to during a model outbreak. This object is used
       to hold the 'FILE' objects for open files, and will
       ensure that these files are closed and written to disk
       as needed. It will also ensure that files are written
       to the correct output directory, and that they are only
       opened when they are needed (e.g. only the first call
       to open the file will actually open it - subsequent
       calls will return the already-open file handler)

       Examples
       --------
       >>> output = OutputFiles(output_dir="output", check_empty=True)
       >>> FILE = output.open("output.txt")
       >>> FILE.write("some output\\n")
       >>> FILE = output.open("something.csv.bz2", auto_bzip=True)
       >>> FILE.write("something,else,is,here\\n")
       >>> output.flush()
       >>> FILE = output.open("output.txt")
       >>> FILE.write("some more output\\n")
       >>> output.close()

       Note that you can also use OutputFiles in a contexthandler, to
       ensure that all output files are automatically closed, e.g.

       >>> with OutputFiles(output_dir="output") as output:
       >>>     FILE = output.open("output.txt")
       >>>     FILE.write("something\\n")
    """

    def __init__(self, output_dir: str = "output",
                 check_empty: bool = True,
                 force_empty: bool = False,
                 prompt=input,
                 auto_bzip: bool = False):
        """Construct a set of OutputFiles. These will all be written
           to 'output_dir'.

           Parameters
           ----------
           output_dir: str
             The directory in which to create all of the output files.
             This directory will be created automatically if it doesn't
             exist
           check_empty: bool
             Whether or not to check if the directory is empty before
             continuing. If the directory is not empty, then the user
             will be prompted to make a decision to either keep going,
             choose a different directory, remove existing output
             or exit
           force_empty: bool
             Force the output directory to be empty. BE CAREFUL as this
             will remove all files in that directory! There are checks
             to stop you doing something silly, but these are not
             fool-proof. The user will be prompted to confirm that
             the files should be removed
           prompt:
             This is the function that should be called to prompt the
             user for input, e.g. to confirm whether or not files
             should be deleted. This defaults to `input`. Set this
             to None if you *really* want MetaWards to remove files
             silently (e.g. useful if you are running batch jobs
             on a cluster and you really know what you are doing)
           auto_bzip: bool
             The default flag for `auto_bzip` when opening files. If
             this is true then all files will be automatically bzipped
             (compressed) as they are written, unless the code opening
             the file has explicitly asked otherwise
        """
        self._check_empty = _get_bool(check_empty)
        self._force_empty = _get_bool(force_empty)
        self._auto_bzip = _get_bool(auto_bzip)
        self._prompt = prompt
        self._output_dir = output_dir
        self._is_open = False
        self._open_files = {}
        self._filenames = {}

        self._open_dir()

    def __enter__(self):
        self._open_dir()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._close_dir()

    def _open_dir(self):
        """Internal function used to open the directory in which
           all output files will be placed
        """
        if self._is_open or self._output_dir is None:
            return

        import os

        if self._output_dir is None:
            raise ValueError("You cannot open an empty OutputFiles!")

        outdir = _expand(self._output_dir)

        if os.path.exists(outdir):
            outdir = os.path.abspath(outdir)

            if self._check_empty:
                if not _is_empty(outdir):
                    if self._force_empty:
                        _force_remove(outdir, self._prompt)
                    else:
                        _check_remove(outdir, self._prompt)

                    # remake the directory after it has been removed
                    try:
                        os.makedirs(outdir)
                    except FileExistsError:
                        pass

            if not os.path.isdir(outdir):
                from .utils._console import Console
                Console.error(
                    f"Cannot open {outdir} as it is not a directory!")
                raise FileExistsError(f"{outdir} is an existing file!")

        try:
            os.makedirs(outdir)
        except FileExistsError as e:
            # this is no problem, as we have already validated
            # that the directory already existing is ok
            if not os.path.isdir(outdir):
                # but it is a problem if it is not a directory...
                raise e

        self._output_dir = str(_Path(outdir).absolute().resolve())

        self._is_open = True

    def _close_dir(self):
        """Internal function used to close all of the output files"""
        if not self._is_open:
            return

        errors = []

        for filename, handle in self._open_files.items():
            try:
                handle.close()
            except Exception as e:
                errors.append(f"Could not close {filename}: "
                              f"{e.__class__} {e}")

        self._is_open = False
        self._open_files = {}
        self._filenames = {}

    @staticmethod
    def remove(path, prompt=input):
        """Remove the passed filename or directory

           Parameters
           ----------
           path: str
             The path to the file or directory or remove
           prompt:
             Prompt to use to ask the user - if None then no checks!
        """
        path = _Path(_expand(path)).absolute().resolve()
        _force_remove(path, prompt)

    def is_open(self):
        """Return whether or not the output files are open"""
        return self._is_open

    def is_closed(self):
        """Return whether or not the output files are closed"""
        return not self.is_open()

    def output_dir(self):
        """Return the absolute path of the directory to which
           the output files will be written
        """
        return self._output_dir

    def open(self, filename: str, auto_bzip=None, mode="t",
             headers=None, sep=" "):
        """Open the file called 'filename' in the output directory,
           returning a handle to that file. Note that this will
           open the file once, and will return the already-open
           file handle on all subsequent calls.

           Parameters
           ----------
           filename: str
             The name of the file to open. This must be relative
             to the output directory, and within that directory.
             It is an error to try to open a file that is
             not contained within this directory.
           auto_bzip: bool
             Whether or not to open the file in auto-bzip (compression)
             mode. If this is True then the file will be automatically
             compressed as it is written. The filename will have
             '.bz2' automatically appended so that this is clear.
             If this is False then the file will be written uncompressed.
             If 'None' is passed (the default) then the value of
             `auto_bzip` that was passed to the constructor of
             this OutputFiles will be used. Note that this flag is
             ignored if the file is already open.
           mode: str
             The mode of opening the file, e.g. 't' for text mode, and
             'b' for binary mode. The default is text mode
           headers: list[str] or plain str
             The headers to add to the top of the file, e.g. if it will
             contain column data. This will be written to the first line
             when the file is opened. If a list is passed, then this
             will be written joined together using 'sep'. If a plain
             string is passed then this will be written.
             If nothing is passed then no headers will be written.
           sep: str
             The separator used for the headers (e.g. " " or "," are good
             choices). By default things are space-separated

           Returns
           -------
           file
             The handle to the open file
        """
        import os

        self._open_dir()

        outdir = self._output_dir
        p = _Path(_expand(filename))

        if not p.is_absolute():
            p = _Path(os.path.join(outdir, filename))

        filename = str(p.absolute().resolve())

        prefix = os.path.commonprefix([outdir, filename])
        if prefix != outdir:
            raise ValueError(f"You cannot try to open {filename} as "
                             f"this is not in the output directory "
                             f"{outdir} - common prefix is {prefix}")

        if filename in self._open_files:
            return self._open_files[filename]

        if auto_bzip is None:
            auto_bzip = self._auto_bzip

        auto_bzip = _get_bool(auto_bzip)

        if mode is None:
            mode = "w"
        elif mode.find("w") == -1:
            mode = f"w{mode}"

        if mode.find("b") == -1:
            # text file = encoding should be "UTF-8"
            encoding = "UTF-8"
        else:
            encoding = None

        if auto_bzip:
            import bz2
            if not filename.endswith(".bz2"):
                suffix = ".bz2"
            else:
                suffix = ""

            if encoding:
                FILE = bz2.open(f"{filename}{suffix}", mode=mode,
                                encoding=encoding)
            else:
                FILE = bz2.open(f"{filename}{suffix}", mode=mode)

            self._open_files[filename] = FILE
            self._filenames[filename] = f"{filename}{suffix}"
        else:
            if encoding:
                FILE = open(filename, mode=mode, encoding=encoding)
            else:
                FILE = open(filename, mode=mode)

            self._open_files[filename] = FILE
            self._filenames[filename] = filename

        if headers is not None:
            if isinstance(headers, str):
                FILE.write(headers)
                FILE.write("\n")
            else:
                FILE.write(sep.join([str(x) for x in headers]))
                FILE.write("\n")

        return FILE

    def open_subdir(self, dirname):
        """Create and open a sub-directory in this OutputFiles
           called 'dirname'. This will inherit all properties,
           e.g. check_empty, auto_bzip etc from this OutputFiles

           Parameters
           ----------
           dirname: str
             The name of the subdirectory to open

           Returns
           -------
           subdir: OutputFiles
             The open subdirectory
        """
        import os

        self._open_dir()

        outdir = self._output_dir
        p = _Path(_expand(dirname))

        if not p.is_absolute():
            p = _Path(os.path.join(outdir, dirname))

        subdir = str(p.absolute().resolve())

        prefix = os.path.commonprefix([outdir, subdir])
        if prefix != outdir:
            raise ValueError(f"You cannot try to open {dirname} as "
                             f"this is not in the output directory "
                             f"{outdir} - common prefix is {prefix}")

        return OutputFiles(output_dir=subdir, check_empty=self._check_empty,
                           force_empty=self._force_empty, prompt=self._prompt,
                           auto_bzip=self._auto_bzip)

    def auto_bzip(self):
        """Return whether the default is to automatically bzip2 files"""
        return self._auto_bzip

    def get_path(self):
        """Return the full expanded path to this directory"""
        return self._output_dir

    def get_filename(self, filename):
        """Return the full expanded filename for 'filename'"""
        import os

        self._open_dir()

        outdir = self._output_dir
        p = _Path(_expand(filename))

        if not p.is_absolute():
            p = _Path(os.path.join(outdir, filename))

        filename = str(p.absolute().resolve())

        if filename in self._filenames:
            return self._filenames[filename]
        else:
            raise FileNotFoundError(f"No open file {filename}")

    def close(self):
        """Close all of the files and this directory"""
        self._close_dir()

    def flush(self):
        """Flush the contents of all files to disk"""
        for filename, handle in self._open_files.items():
            try:
                handle.flush()
            except Exception:
                pass




from typing import Union as _Union

__all__ = ["safe_eval_number"]


def safe_eval_number(s: _Union[float, int, str]) -> float:
    """Convert 's' to a number. This supports normal floats,
       but also simple maths expressions like 1/1.2,
       plus anything that ends with a "%" is recognised
       as a percentage

       Examples
       --------

       safe_eval_number(0.3)         -> 0.3
       safe_eval_number("5%")        -> 0.05
       safe_eval_number("1/4")       -> 0.25
       safe_eval_number("(30+100)%)  -> 1.3
    """
    if isinstance(s, float) or isinstance(s, int):
        return s

    if not isinstance(s, str):
        s = str(s)

    try:
        # this is about as save as eval gets in python
        x = eval(s, {"__builtins__": None}, {})

        v = float(x)
        if v.is_integer():
            return int(x)
        else:
            return v
    except Exception:
        pass

    if isinstance(s, str):
        s = s.strip()
        if s.endswith("%"):
            try:
                return safe_eval_number(s[0:-1]) / 100.0
            except Exception:
                pass

    raise ValueError(f"Cannot interpret '{s}' as a float")




from typing import Union as _Union
from typing import List as _List

cimport cython
from cython.parallel import parallel, prange
from libc.stdint cimport uintptr_t

from libc.math cimport ceil

from .._networks import Networks
from .._infections import Infections
from .._demographics import DemographicID, DemographicIDs

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["go_isolate", "go_isolate_serial", "go_isolate_parallel"]


def go_isolate_parallel(go_from: _Union[DemographicID, DemographicIDs],
                        go_to: DemographicID,
                        network: Networks,
                        infections: Infections,
                        nthreads: int,
                        rngs,
                        self_isolate_stage: _Union[_List[int], int] = 2,
                        fraction: _Union[_List[float], float] = 1.0,
                        **kwargs) -> None:
    """This go function will move individuals from the "from"
       demographic(s) to the "to" demographic if they show any
       signs of infection (the disease stage is greater or equal
       to 'self_isolate_stage' - by default this is level '2',
       which is one level above "latent"). This can move
       a subset of individuals if 'fraction' is less than 1, e.g.
       0.5 would move 50% of individuals (chosen using
       a random binomial distribution)

       Parameters
       ----------
       from: DemographicID or DemographicIDs
         ID (name or index) or IDs of the demographics to scan for
         infected individuals
       to: DemographicID
         ID (name or index) of the isolation demographic to send infected
         individuals to
       network: Networks
         The networks to be modelled. This must contain all of the
         demographics that are needed for this go function
       self_isolate_stage: int or List[int]
         The stage of infection an individual must be at before they
         are moved into this demographic. If a list is passed then
         this can be multiple stages, e.g. [2, 3] will move at
         stages 2 and 3. Multiple stages are needed if only
         a fraction of individuals move.
       fraction: float or List[float]
         The fraction (percentage) of individuals who are moved from
         this stage into isolation. If this is a single value then
         the same fraction applies to all self_isolation_stages. Otherwise,
         the fraction for self_isolate_stage[i] is fraction[i]
       rngs:
         Thread-safe random number generators used to choose the fraction
         of individuals
    """

    # make sure that all of the needed demographics exist, and
    # convert them into a canonical form (indexes, list[indexes])
    if not isinstance(go_from, list):
        go_from = [go_from]

    subnets = network.subnets
    demographics = network.demographics
    subinfs = infections.subinfs

    go_from = [demographics.get_index(x) for x in go_from]

    go_to = demographics.get_index(go_to)

    to_subnet = subnets[go_to]
    to_subinf = subinfs[go_to]

    if isinstance(self_isolate_stage, list):
        stages = [int(stage) for stage in self_isolate_stage]
    else:
        stages = [int(self_isolate_stage)]

    if isinstance(fraction, list):
        fractions = [float(frac) for frac in fraction]
    else:
        fractions = [float(fraction)] * len(stages)

    for fraction in fractions:
        if fraction < 0 or fraction > 1:
            raise ValueError(
                f"The move fractions {fractions} should all be 0 to 1")

    N_INF_CLASSES = infections.N_INF_CLASSES

    for stage in stages:
        if stage < 1 or stage >= N_INF_CLASSES:
            raise ValueError(
                f"The stage(s) of self-isolation {stages} "
                f"is invalid for a disease with {N_INF_CLASSES} stages")

    if len(stages) != len(fractions):
        raise ValueError(
            f"The number of self isolation stages {stages} must equal "
            f"the number of fractions {fractions}")

    cdef int nnodes_plus_one = 0
    cdef int nlinks_plus_one = 0

    cdef int to_move = 0

    cdef int * work_infections_i
    cdef int * play_infections_i

    cdef int * to_work_infections_i
    cdef int * to_play_infections_i

    cdef int nsubnets = len(subnets)

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    cdef int thread_id

    cdef int num_threads = nthreads

    cdef int ii = 0
    cdef int i = 0
    cdef int j = 0

    cdef double fraction_i = 1.0

    for ii in go_from:
        subnet = subnets[ii]
        subinf = subinfs[ii]
        nnodes_plus_one = subinf.nnodes + 1
        nlinks_plus_one = subinf.nlinks + 1

        for i, fraction in zip(stages, fractions):
            fraction_i = fraction

            work_infections_i = get_int_array_ptr(subinf.work[i])
            play_infections_i = get_int_array_ptr(subinf.play[i])

            to_work_infections_i = get_int_array_ptr(to_subinf.work[i])
            to_play_infections_i = get_int_array_ptr(to_subinf.play[i])

            with nogil, parallel(num_threads=num_threads):
                if fraction_i == 1.0:
                    for j in prange(1, nlinks_plus_one, schedule="static"):
                        if work_infections_i[j] > 0:

                            to_work_infections_i[j] = \
                                                to_work_infections_i[j] + \
                                                work_infections_i[j]
                            work_infections_i[j] = 0

                    for j in prange(1, nnodes_plus_one, schedule="static"):
                        if play_infections_i[j] > 0:
                            to_play_infections_i[j] = \
                                                to_play_infections_i[j] + \
                                                play_infections_i[j]
                            play_infections_i[j] = 0
                else:
                    thread_id = cython.parallel.threadid()
                    rng = _get_binomial_ptr(rngs_view[thread_id])

                    for j in prange(1, nlinks_plus_one, schedule="static"):
                        to_move = _ran_binomial(rng, fraction_i,
                                                <int>(work_infections_i[j]))

                        if to_move > 0:
                            to_work_infections_i[j] = \
                                                to_work_infections_i[j] + \
                                                to_move
                            work_infections_i[j] = \
                                                work_infections_i[j] - \
                                                to_move

                    for j in prange(1, nnodes_plus_one, schedule="static"):
                        to_move = _ran_binomial(rng, fraction_i,
                                                <int>(play_infections_i[j]))

                        if to_move > 0:
                            to_play_infections_i[j] = \
                                                to_play_infections_i[j] + \
                                                to_move
                            play_infections_i[j] = \
                                                play_infections_i[j] - \
                                                to_move


def go_isolate_serial(go_from: _Union[DemographicID, DemographicIDs],
                      go_to: DemographicID,
                      network: Networks,
                      infections: Infections,
                      rngs,
                      self_isolate_stage: _Union[_List[int], int] = 2,
                      fraction: _Union[_List[float], float] = 1.0,
                      **kwargs) -> None:
    """This go function will move individuals from the "from"
       demographic(s) to the "to" demographic if they show any
       signs of infection (the disease stage is greater or equal
       to 'self_isolate_stage' - by default this is level '2',
       which is one level above "latent"). This can move
       a subset of individuals if 'fraction' is less than 1, e.g.
       0.5 would move 50% of individuals (chosen using
       a random binomial distribution)

       Parameters
       ----------
       from: DemographicID or DemographicIDs
         ID (name or index) or IDs of the demographics to scan for
         infected individuals
       to: DemographicID
         ID (name or index) of the demographic to send infected
         individuals to
       network: Networks
         The networks to be modelled. This must contain all of the
         demographics that are needed for this go function
       self_isolate_stage: int or List[int]
         The stage of infection an individual must be at before they
         are moved into this demographic. If a list is passed then
         this can be multiple stages, e.g. [2, 3] will move at
         stages 2 and 3. Multiple stages are needed if only
         a fraction of individuals move.
       fraction: float or List[float]
         The fraction (percentage) of individuals who are moved from
         this stage into isolation. If this is a single value then
         the same fraction applies to all self_isolation_stages. Otherwise,
         the fraction for self_isolate_stage[i] is fraction[i]
       rngs:
         Thread-safe random number generators used to choose the fraction
         of individuals
    """

    # make sure that all of the needed demographics exist, and
    # convert them into a canonical form (indexes, list[indexes])
    if not isinstance(go_from, list):
        go_from = [go_from]

    subnets = network.subnets
    demographics = network.demographics
    subinfs = infections.subinfs

    go_from = [demographics.get_index(x) for x in go_from]

    go_to = demographics.get_index(go_to)

    to_subnet = subnets[go_to]
    to_subinf = subinfs[go_to]

    if isinstance(self_isolate_stage, list):
        stages = [int(stage) for stage in self_isolate_stage]
    else:
        stages = [int(self_isolate_stage)]

    if isinstance(fraction, list):
        fractions = [float(frac) for frac in fraction]
    else:
        fractions = [float(fraction)] * len(stages)

    for fraction in fractions:
        if fraction < 0 or fraction > 1:
            raise ValueError(
                f"The move fractions {fractions} should all be 0 to 1")

    N_INF_CLASSES = infections.N_INF_CLASSES

    for stage in stages:
        if stage < 1 or stage >= N_INF_CLASSES:
            raise ValueError(
                f"The stage(s) of self-isolation {stages} "
                f"is invalid for a disease with {N_INF_CLASSES} stages")

    if len(stages) != len(fractions):
        raise ValueError(
            f"The number of self isolation stages {stages} must equal "
            f"the number of fractions {fractions}")

    cdef int nnodes_plus_one = 0
    cdef int nlinks_plus_one = 0

    cdef int to_move = 0

    cdef int * work_infections_i
    cdef int * play_infections_i

    cdef int * to_work_infections_i
    cdef int * to_play_infections_i

    cdef int nsubnets = len(subnets)

    cdef int ii = 0
    cdef int i = 0
    cdef int j = 0

    # get the random number generator
    cdef binomial_rng* rng = _get_binomial_ptr(rngs[0])

    cdef double fraction_i = 1.0

    for ii in go_from:
        subnet = subnets[ii]
        subinf = subinfs[ii]
        nnodes_plus_one = subinf.nnodes + 1
        nlinks_plus_one = subinf.nlinks + 1

        for i, fraction in zip(stages, fractions):
            fraction_i = fraction

            work_infections_i = get_int_array_ptr(subinf.work[i])
            play_infections_i = get_int_array_ptr(subinf.play[i])

            to_work_infections_i = get_int_array_ptr(to_subinf.work[i])
            to_play_infections_i = get_int_array_ptr(to_subinf.play[i])

            if fraction_i == 1.0:
                for j in range(1, nlinks_plus_one):
                    if work_infections_i[j] > 0:
                        to_work_infections_i[j] = to_work_infections_i[j] + \
                                                  work_infections_i[j]
                        work_infections_i[j] = 0

                for j in range(1, nnodes_plus_one):
                    if play_infections_i[j] > 0:
                        to_play_infections_i[j] = to_play_infections_i[j] + \
                                                  play_infections_i[j]
                        play_infections_i[j] = 0
            else:
                for j in range(1, nlinks_plus_one):
                    to_move = _ran_binomial(rng, fraction_i,
                                            <int>(work_infections_i[j]))
                    if to_move > 0:
                        to_work_infections_i[j] = to_work_infections_i[j] + \
                                                  to_move
                        work_infections_i[j] = work_infections_i[j] - \
                                               to_move

                for j in range(1, nnodes_plus_one):
                    to_move = _ran_binomial(rng, fraction_i,
                                            <int>(play_infections_i[j]))
                    if to_move > 0:
                        to_play_infections_i[j] = to_play_infections_i[j] + \
                                                  to_move
                        play_infections_i[j] = play_infections_i[j] - \
                                               to_move


def go_isolate(nthreads: int = 1, **kwargs) -> None:
    """This go function will move individuals from the "from"
       demographic(s) to the "to" demographic if they show any
       signs of infection (the disease stage is greater or equal
       to 'self_isolate_stage' - by default this is level '2',
       which is one level above "latent"). Individuals are held
       in the new demographic for "duration" days, before being
       returned either to their original demographic, or
       released to the "release_to" demographic

       Parameters
       ----------
       from: DemographicID or DemographicIDs
         ID (name or index) or IDs of the demographics to scan for
         infected individuals
       to: DemographicID
         ID (name or index) of the demographic to send infected
         individuals to
       network: Networks
         The networks to be modelled. This must contain all of the
         demographics that are needed for this go function
       self_isolate_stage: int
         The stage of infection an individual must be at before they
         are moved into this demographic
       duration: int
         The number of days an individual should isolate for
       release_to: DemographicID
         ID (name or index) that the individual should move to after
         existing isolation. If this is not set, then the individual
         will return to their original demographic
    """

    if nthreads > 1:
        go_isolate_parallel(nthreads=nthreads, **kwargs)
    else:
        go_isolate_serial(**kwargs)



"""
.. currentmodule:: metawards.app

Functions
=========

.. autosummary::
    :toctree: generated/

    get_parallel_scheme
    parse_args
    get_hostfile
    get_cores_per_node
    get_threads_per_task
    scoop_supervisor
    mpi_supervisor
    cli

"""

from .run import *
from .plot import *



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
cimport openmp

from .._network import Network
from .._node import Node
from ._profiler import Profiler

from ._get_array_ptr cimport get_double_array_ptr

__all__ = ["reset_work_matrix", "reset_play_matrix",
           "reset_play_susceptibles", "reset_everything"]


def reset_work_matrix(network: Network, nthreads=1):
    """Resets the work entries in the passed Network. You need
       to call this before performing a new model run

       Parameters
       ----------
       network: Network
         The network to be reset
       nthreads: int
         Number of threads to parallelise over
    """
    links = network.links

    cdef int i = 0
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef double * links_weight = get_double_array_ptr(links.weight)

    cdef int nlinks_plus_one = network.nlinks + 1
    cdef int num_threads = nthreads

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nlinks_plus_one, schedule="static"):
            links_suscept[i] = links_weight[i]


def reset_play_matrix(network: Network, nthreads: int=1):
    """Resets the play entries in the passed Network. You need
       to call this before performing a new model run

       Parameters
       ----------
       network: Network
         The network to be reset
       nthreads: int
         Number of threads to parallelise over
    """
    links = network.play

    cdef int i = 0
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef double * links_weight = get_double_array_ptr(links.weight)

    cdef int nlinks_plus_one = network.nplay + 1
    cdef int num_threads = nthreads

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nlinks_plus_one, schedule="static"):
            links_weight[i] = links_suscept[i]


def reset_play_susceptibles(network: Network, nthreads: int=1):
    """Resets the ward entries in the passed Network. You need
       to call this before performing a new model run

       Parameters
       ----------
       network: Network
         The network to be reset
       nthreads: int
         Number of threads to parallelise over
    """
    nodes = network.nodes

    cdef int i = 0
    cdef double * nodes_play_suscept = get_double_array_ptr(nodes.play_suscept)
    cdef double * nodes_save_play_suscept = get_double_array_ptr(
                                                    nodes.save_play_suscept)

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int num_threads = nthreads

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nnodes_plus_one, schedule="static"):
            nodes_play_suscept[i] = nodes_save_play_suscept[i]


def reset_everything(network: Network, profiler: Profiler, nthreads: int = 1):
    """Reset everything in the passed network so that it can
       be used for a new model run

       Parameters
       ----------
       network: Network
         The network to be reset
       nthreads: int
         Number of threads to use during the reset
       profiler: Profiler
         Profiler used to profile the reset
    """
    p = profiler.start("reset_work")
    reset_work_matrix(network=network, nthreads=nthreads)
    p = p.stop()

    p = p.start("reset_play")
    reset_play_matrix(network=network, nthreads=nthreads)
    p = p.stop()

    p = p.start("reset_susceptibles")
    reset_play_susceptibles(network=network, nthreads=nthreads)
    p = p.stop()

    p = p.start("reset_params")
    params = network.params
    if params:
        N_INF_CLASSES = params.disease_params.N_INF_CLASSES()

        params.disease_params.contrib_foi = N_INF_CLASSES * [0]

        for i in range(0, N_INF_CLASSES-1):   # why -1?
            params.disease_params.contrib_foi[i] = 1

        network.params = params

    p = p.stop()



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

from libc.stdio cimport FILE, fopen, fscanf, fclose, feof

from .._network import Network
from .._links import Links
from ._profiler import Profiler, NullProfiler

from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["build_play_matrix"]

def build_play_matrix(network: Network,
                      max_nodes: int, max_links: int,
                      profiler: Profiler=None):
    """Build the play matrix for the passed network"""
    if profiler is None:
        profiler = NullProfiler()

    p = profiler.start("build_play_matrix")

    params = network.params

    if params is None:
        return

    nodes = network.nodes

    p = p.start("allocate_memory")
    links = Links(network.max_links + 1)
    p = p.stop()

    cdef int nnodes_plus_one = network.nnodes + 1

    cdef int nlinks = 0
    cdef int j = 0
    cdef int from_id = 0
    cdef int to_id = 0
    cdef double weight = 0.0

    cdef int * nodes_label = get_int_array_ptr(nodes.label)
    cdef int * nodes_begin_p = get_int_array_ptr(nodes.begin_p)
    cdef int * nodes_end_p = get_int_array_ptr(nodes.end_p)
    cdef int * nodes_self_p = get_int_array_ptr(nodes.self_p)

    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)
    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)

    cdef double * nodes_denominator_p = get_double_array_ptr(
                                                    nodes.denominator_p)
    cdef double * nodes_play_suscept = get_double_array_ptr(
                                                    nodes.play_suscept)

    cdef int error_from = -1
    cdef int error_to = -1

    cdef int MAX_LINKS = max_links
    cdef int MAX_NODES = max_nodes

    cdef char* fname
    cdef FILE* cfile

    from ._console import Console

    # need to use C file reading as too slow in python
    if params.input_files.play is None:
        Console.print("No play links file to read")
    else:
        p = p.start("read_play_file")
        filename = params.input_files.play.encode("UTF-8")
        fname = filename
        cfile = fopen(fname, "r")

        if cfile == NULL:
            raise FileNotFoundError(f"No such file or directory: {filename}")

        # resets the node label as a flag to check progress?
        with nogil:
            for j in range(1, nnodes_plus_one):
                nodes_label[j] = -1

            while not feof(cfile):
                fscanf(cfile, "%d %d %lf\n", &from_id, &to_id, &weight)

                nlinks += 1

                if nlinks >= MAX_LINKS:
                    break

                if from_id == 0 or to_id == 0:
                    error_from = from_id
                    error_to = to_id
                    break

                if nodes_label[from_id] == -1:
                    nodes_label[from_id] = from_id
                    nodes_begin_p[from_id] = nlinks
                    nodes_end_p[from_id] = nlinks

                if from_id == to_id:
                    nodes_self_p[from_id] = nlinks

                nodes_end_p[from_id] += 1

                links_ifrom[nlinks] = from_id
                links_ito[nlinks] = to_id
                links_weight[nlinks] = weight

                nodes_denominator_p[from_id] += weight
                nodes_play_suscept[from_id] += weight

            fclose(cfile)

        if nlinks >= MAX_LINKS:
            raise MemoryError(f"There are too many links (>{nlinks}) to fit "
                              f"into pre-allocated memory (max_links = "
                              f"{max_links}). Increase this and try again.")

        if error_from != -1 or error_to != -1:
            raise ValueError(f"{params.input_files.play} is corrupted. "
                             f"Zero in link list: ${error_from}-${error_to}! "
                             f"Renumber files and start again")

        p = p.stop()
    # end of if have playfile

    p = p.start("renormalise?")
    renormalise = (params.input_files.play == params.input_files.work)
    p = p.stop()

    if renormalise:
        Console.print("Identical work and play links, so renormalising...")

    p = p.start("renormalise_loop")
    for j in range(1, nlinks+1):   # careful 1-indexed
        if renormalise:
            links_weight[j] /= nodes_denominator_p[links_ifrom[j]]

        links_suscept[j] = links_weight[j]
    p = p.stop()

    from . import fill_in_gaps
    p = p.start("fill_in_gaps")
    fill_in_gaps(network, max_nodes=max_nodes)
    p = p.stop()

    cdef int i1 = 0
    cdef int i2 = 0
    cdef double [::1] nodes_save_play_suscept = nodes.save_play_suscept

    cdef int max_node_id = network.nnodes

    if params.input_files.play_size is None:
        Console.print("No play_size file to read")
    else:
        # need to use C file reading as too slow in python
        p = p.start("read_play_size_file")
        filename = params.input_files.play_size.encode("UTF-8")
        fname = filename
        cfile = fopen(fname, "r")

        if cfile == NULL:
            raise FileNotFoundError(f"No such file or directory: {filename}")

        try:
            while not feof(cfile):
                fscanf(cfile, "%d %d\n", &i1, &i2)

                if i1 > max_node_id:
                    max_node_id = i1

                nodes_play_suscept[i1] = i2
                nodes_denominator_p[i1] = i2
                nodes_save_play_suscept[i1] = i2

            fclose(cfile)
        except Exception as e:
            fclose(cfile)
            raise ValueError(f"{params.input_files.play_size} is corrupted or "
                            f"unreadable? Error = {e.__class__}: {e}")

        # we now need to fill in the missing nodes that are defined
        # in the play_size file, but were not linked to in the node
        # links file
        old_nnodes = network.nnodes

        if max_node_id >= max_nodes:
            raise MemoryError(f"Link ID {max_node_id} implies we have more "
                              f"nodes than are pre-allocated ({max_nodes}). "
                              f"Increase this and try again.")

        Console.print(
            f"Adding missing nodes from {old_nnodes+1} to {max_node_id}")

        network.nnodes = max_node_id

        for i in range(old_nnodes, max_node_id+1):
            network.nodes.label[i] = i

        p = p.stop()
    # end of if play_size file

    Console.print(f"Number of play links equals {nlinks}")

    network.nplay = nlinks
    network.play = links

    p.stop()




from ._link import Link


__all__ = ["Links"]


class Links:
    """This is a container class for Links. This uses arrays
       to store a list of Link objects as a "struct of arrays".
       This should improve speed of loading and access.
    """

    def __init__(self, N: int = 0):
        """Create a container for up to "N" Links"""
        if N <= 0:
            self._is_null = True
        else:
            self._is_null = False

            from .utils._array import create_double_array, create_int_array

            # Struct of arrays for each piece of data. See the
            # Link class for information about what each variable
            # holds. Code uses "-1" to represent a null value
            self.ifrom = create_int_array(N, -1)
            self.ito = create_int_array(N, -1)

            self.weight = create_double_array(N, 0.0)
            self.suscept = create_double_array(N, 0.0)
            self.distance = create_double_array(N, 0.0)

    def is_null(self):
        return self._is_null

    def __len__(self):
        """Return the number of links in this container"""
        if self._is_null:
            return 0
        else:
            return len(self.ifrom)

    def population(self) -> int:
        """Return the population of these links"""
        if self.suscept is None:
            return 0
        else:
            return sum(self.suscept)

    def copy(self):
        """Return a copy of these links, using a shallow copy for
           things that stay the same (e.g. ifrom, ito, distance)
           and a deep copy for things that are variable
           (e.g. weight and suscept)
        """
        from copy import copy, deepcopy
        links = copy(self)

        links.weight = deepcopy(self.weight)
        links.suscept = deepcopy(self.suscept)

        return links

    def assert_not_null(self):
        """Assert that this collection of links is not null"""
        assert not self.is_null()

    def assert_valid_index(self, i):
        """Assert that the passed index 'i' is valid for this collection.
           This will return the index with Python reverse indexing,
           so that "-1" refers to the last link in the collection
        """
        self.assert_not_null()
        n = len(self)

        if i < 0:
            i = n + i

        if i < 0 or i >= n:
            raise IndexError(f"Invalid link index {i}. Number of links "
                             f"in this container equals {n}")

        return i

    def __getitem__(self, i: int):
        """Return the link at index 'i'. Note that this is a
           deep copy of the link. Changing the link will not change
           the data in this container. To update the data in this
           container you need to use __setitem__, e.g. via the
           index operator (see __setitem__)
        """
        i = self.assert_valid_index(i)

        return Link(ifrom=self.ifrom[i], ito=self.ito[i],
                    weight=self.weight[i], suscept=self.suscept[i],
                    distance=self.distance[i])

    def __setitem__(self, i: int, value: Link):
        """Set the item as index 'i' equal to 'value'. This deep-copies
           'value' into this container
        """
        i = self.assert_valid_index(i)

        self.ifrom[i] = value.ifrom if value.ifrom is not None else -1
        self.ito[i] = value.ito if value.ito is not None else -1
        self.weight[i] = value.weight if value.weight is not None else 0.0
        self.suscept[i] = value.suscept if value.suscept is not None else 0.0
        self.distance[i] = value.distance if value.distance is not None else 0.0

    def resize(self, N: int):
        """Resize this container to hold 'N' links. This will expand
           the container if 'N' is greater than len(self), or will
           contract the container (deleting excess links) if 'N'
           is less than len(self). This function is called typically
           when you pre-allocate a large Links container, and then
           want to reduce the size to fit the number of loaded links
        """
        if self.is_null():
            # Assign to a new Links object created with this size
            self.__dict__ = Links(N).__dict__
            return
        elif N <= 0:
            self.__dict__ = Links(0).__dict__

        size = len(self)

        if N == size:
            return

        from .utils._array import resize_array

        self.ifrom = resize_array(self.ifrom, N, -1)
        self.ito = resize_array(self.ito, N, -1)
        self.weight = resize_array(self.weight, N, 0.0)
        self.suscept = resize_array(self.suscept, N, 0.0)
        self.distance = resize_array(self.distance, N, 0.0)

    def scale_susceptibles(self, ratio: any = None):
        """Scale the number of susceptibles in these Links
           by the passed scale ratio. This can be a value, e.g.
           ratio = 2.0 will scale the total number of susceptibles
           by 2.0. This can also be lists of values,
           where ward[i] will be scaled by ratio[i]. They can also
           be dictionaries, e.g. ward[i] scaled by ratio[i]

           Parameters
           ----------
           ratio: None, float, list or dict
             The amount by which to scale the total population of
             susceptibles - evenly scales the work and play populations

           Returns
           -------
           None
        """
        from .utils._scale_susceptibles import scale_link_susceptibles
        scale_link_susceptibles(links=self, ratio=ratio)




from typing import Union as _Union
from typing import List as _List
from typing import IO as _IO

from contextlib import contextmanager as _contextmanager


__all__ = ["Console"]


# Global rich.Console()
_console = None

# Global console theme
_theme = None


class _NullSpinner:
    """Null spinner to use if yaspin isn't available"""

    def __init__(self):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return self

    def success(self):
        pass

    def failure(self):
        pass


class Console:
    """This is a singleton class that provides access to printing
       and logging functions to the console. This uses 'rich'
       for rich console printing
    """
    @staticmethod
    def supports_emojis():
        """Return whether or not you can print emojis to this console"""
        import sys
        if sys.platform == "win32":
            return False
        else:
            return True

    @staticmethod
    def set_debugging_enabled(enabled, level=None):
        """Switch on or off debugging output"""
        console = Console._get_console()
        console._debugging_enabled = bool(enabled)
        console._debugging_level = level

    @staticmethod
    def set_theme(theme):
        """Set the theme used for the console - this should be
           one of the themes in metawards.themes
        """
        global _theme

        if isinstance(theme, str):
            if theme.lower().strip() == "simple":
                from metawards.themes import Simple
                _theme = Simple()
            elif theme.lower().strip() == "default":
                from metawards.themes import SpringFlowers
                _theme = SpringFlowers()
        else:
            _theme = theme

    @staticmethod
    def _get_theme():
        global _theme

        if _theme is None:
            from metawards.themes import SpringFlowers
            _theme = SpringFlowers()

        return _theme

    @staticmethod
    def _get_console():
        global _console

        if _console is None:
            from rich.console import Console as _Console
            theme = Console._get_theme()

            _console = _Console(record=True,
                                highlight=theme.should_highlight(),
                                highlighter=theme.highlighter(),
                                markup=theme.should_markup(),
                                log_time=True, log_path=True,
                                emoji=Console.supports_emojis())

            _console._use_spinner = True
            _console._debugging_enabled = False
            _console._debugging_level = None

            # also install pretty traceback support
            from rich.traceback import install as _install_rich
            _install_rich()

        return _console

    @staticmethod
    @_contextmanager
    def redirect_output(outdir: str, auto_bzip: bool = True):
        """Redirect all output and error to the directory 'outdir'"""
        import os as os
        import sys as sys
        import bz2
        from rich.console import Console as _Console

        outfile = os.path.join(outdir, "output.txt")
        errfile = os.path.join(outdir, "output.err")

        if auto_bzip:
            outfile += ".bz2"
            errfile += ".bz2"

            OUTFILE = bz2.open(outfile, "wt", encoding="utf-8")
            ERRFILE = bz2.open(errfile, "wt", encoding="utf-8")
        else:
            OUTFILE = open(outfile, "wt", encoding="utf-8")
            ERRFILE = open(errfile, "wt", encoding="utf-8")

        console = Console._get_console()

        if console is None:
            raise AssertionError("The global console should never be None")

        new_out = _Console(file=OUTFILE, record=False, log_time=True,
                           log_path=True, emoji=Console.supports_emojis())
        new_out._use_spinner = False
        new_out._debugging_enabled = console._debugging_enabled
        new_out._debugging_level = console._debugging_level
        old_out = console

        global _console
        _console = new_out

        new_err = ERRFILE
        old_err = sys.stderr
        sys.stderr = new_err

        try:
            yield new_out
        finally:
            _console = old_out
            sys.stderr = old_err
            OUTFILE.close()
            ERRFILE.close()

    @staticmethod
    def debugging_enabled(level: int = None):
        """Return whether debug output is enabled (optionally for
           the specified level) - if not, then
           anything sent to 'debug' (for that level) is not printed
        """
        console = Console._get_console()

        if not console._debugging_enabled:
            return False

        elif console._debugging_level is not None:
            if level is None:
                return True
            else:
                return level <= console._debugging_level
        else:
            return level is None

    @staticmethod
    def _retrieve_name(variable):
        # thanks to scohe001 on stackoverflow
        # https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string
        import inspect
        callers_local_vars = inspect.currentframe().f_back.f_back.f_locals.items()
        return [var_name for var_name, var_val in callers_local_vars
                if var_val is variable][-1]

    @staticmethod
    def debug(text: str, variables: _List[any] = None, level: int = None,
              markdown: bool = False, **kwargs):
        """Print a debug string to the console. This will only be
           printed if debugging is enabled. You can also print the
           values of variables by passing them as a list to
           'variables'
        """
        if not Console.debugging_enabled(level=level):
            return

        if hasattr(text, "__call__"):
            # the user passed in a lambda for delayed printing
            text = text()

        if not isinstance(text, str):
            text = str(text)

        if level is not None:
            text = f"Level {level}: {text}"

        console = Console._get_console()

        if markdown:
            from rich.markdown import Markdown as _Markdown
            try:
                text = _Markdown(text)
            except Exception:
                text = _Markdown(str(text))

        console.log(text, justify="center", _stack_offset=2, **kwargs)

        if variables is not None:
            from rich.table import Table
            from rich import box
            import inspect
            # get the local variables in the caller's scope
            callers_local_vars = inspect.currentframe().f_back.f_locals.items()

            table = Table(box=box.MINIMAL_DOUBLE_HEAD, style="on magenta")
            table.add_column("Name", justify="right", style="cyan",
                             no_wrap=True)
            table.add_column("Value", justify="left", style="green")

            for variable in variables:
                # get the name of the variable in the caller
                try:
                    # go for the last matching variable in the caller's scope
                    name = [var_name for var_name, var_val
                            in callers_local_vars
                            if var_val is variable][-1]
                except Exception:
                    name = "variable"

                table.add_row(name, str(variable))

            console.print(table)

    @staticmethod
    def print(text: str, markdown: bool = False, style: str = None,
              *args, **kwargs):
        """Print to the console"""
        if markdown:
            from rich.markdown import Markdown as _Markdown
            try:
                text = _Markdown(text)
            except Exception:
                text = _Markdown(str(text))

        theme = Console._get_theme()
        style = theme.text(style)

        try:
            Console._get_console().print(text, style=style)
        except UnicodeEncodeError:
            # this output can't cope with a complex theme - switch
            # to theme 'simple'
            Console.set_theme("simple")

            if isinstance(text, str):
                # try to print this as latin-1
                try:
                    text = text.encode("latin-1",
                                       errors="replace").decode("latin-1")
                    Console._get_console().print(text)
                except Exception:
                    # accept that this isn't going to be printable
                    pass

    @staticmethod
    def rule(title: str = None, style=None, **kwargs):
        """Write a rule across the screen with optional title"""
        from rich.rule import Rule as _Rule
        Console.print("")
        theme = Console._get_theme()
        style = theme.rule(style)
        Console.print(_Rule(title, style=style))

    @staticmethod
    def panel(text: str, markdown: bool = False, width=None,
              padding: bool = True, style: str = None,
              expand=True, *args, **kwargs):
        """Print within a panel to the console"""
        from rich.panel import Panel as _Panel

        if markdown:
            from rich.markdown import Markdown as _Markdown
            text = _Markdown(text)

        theme = Console._get_theme()
        padding_style = theme.padding_style(style)
        style = theme.panel(style)
        box = theme.panel_box(style)

        from rich.padding import Padding as _Padding

        if padding:
            text = _Padding(text, (1, 2), style=padding_style)
        else:
            text = _Padding(text, (0, 1), style=padding_style)

        Console.print(_Panel(text, box=box, width=width,
                             expand=expand,
                             style=style, *args, **kwargs))

    @staticmethod
    def error(text: str, *args, **kwargs):
        """Print an error to the console"""
        Console.rule("ERROR", style="error")
        Console.print(text, style="error", *args, **kwargs)
        Console.rule(style="error")

    @staticmethod
    def warning(text: str, *args, **kwargs):
        """Print a warning to the console"""
        Console.rule("WARNING", style="warning")
        Console.print(text, style="warning", *args, **kwargs)
        Console.rule(style="warning")

    @staticmethod
    def info(text: str, *args, **kwargs):
        """Print an info section to the console"""
        Console.rule("INFO", style="info")
        Console.print(text, style="info", *args, **kwargs)
        Console.rule(style="info")

    @staticmethod
    def center(text: str, *args, **kwargs):
        from rich.text import Text as _Text
        Console.print(_Text(str, justify="center"), *args, **kwargs)

    @staticmethod
    def command(text: str, *args, **kwargs):
        Console.panel(text, style="command", padding=False)

    @staticmethod
    def print_population(population, demographics=None,
                         *args, **kwargs):
        Console.print(population.summary(demographics=demographics))

    @staticmethod
    def print_profiler(profiler, *args, **kwargs):
        Console.print(str(profiler))

    @staticmethod
    def set_use_spinner(use_spinner: bool = True):
        console = Console._get_console()
        console._use_spinner = use_spinner

    @staticmethod
    def spinner(text: str = ""):
        try:
            from yaspin import yaspin, Spinner
            have_yaspin = True
        except ImportError:
            have_yaspin = False

        if not have_yaspin:
            return _NullSpinner()

        console = Console._get_console()

        if console._use_spinner:
            theme = Console._get_theme()
            frames, delay = theme.get_frames(
                width=console.width - len(text) - 10)
            sp = Spinner(frames, delay)

            y = yaspin(sp, text=text, side="right")

            y.success = lambda: theme.spinner_success(y)
            y.failure = lambda: theme.spinner_failure(y)

            return y
        else:
            return _NullSpinner()

    @staticmethod
    def save(file: _Union[str, _IO]):
        """Save the accumulated printing to the console to 'file'.
           This can be a file or a filehandle. The buffer is
           cleared after saving
        """

        # get the console contents
        try:
            text = Console._get_console().export_text(clear=True,
                                                      styles=False)
        except Exception as e:
            Console.error(f"Cannot get console output: {e.__class__} {e}")
            return

        if isinstance(file, str):
            with open(file, "w", encoding="UTF-8") as FILE:
                FILE.write(text)
        else:
            try:
                if file.encoding.lower() != "utf-8":
                    # make sure that encoding the file to the right character
                    # set will replace invalid characters, rather than throw
                    # unicode encoding errors
                    text = text.encode(file.encoding,
                                       errors="replace").decode(file.encoding)

                file.write(text)
            except UnicodeEncodeError:
                # this still didn't work - use a latin-1 round-robin
                # to make everything ascii...
                text = text.encode("latin-1",
                                   errors="replace").decode("latin-1")
                file.write(text)



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
cimport openmp

from .._networks import Networks
from .._infections import Infections
from ._profiler import Profiler

from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["aggregate_networks", "aggregate_infections"]


def aggregate_infections(infections: Infections, profiler: Profiler,
                         nthreads: int=1) -> None:
    """Aggregate all of the demographic sub-network infections data
       together into the overall infections object

       Parameters
       ----------
       infections: Infections
         The complete set of infection data
       profiler: Profiler
         Profiler used to profile this work
       nthreads: int
         Number of threads to use to do this work

       Returns
       -------
       None
    """
    from cython.parallel import parallel, prange

    if profiler is None:
        from ._profiler import NullProfiler
        profiler = NullProfiler()

    p = profiler.start("aggregate_infections")

    cdef int i = 0
    cdef int j = 0
    cdef int idx = 0
    cdef int N_INF_CLASSES = infections.N_INF_CLASSES
    cdef int nnodes_plus_one = infections.nnodes + 1
    cdef int nlinks_plus_one = infections.nlinks + 1
    cdef int nsublinks_plus_one = 0
    cdef int num_threads = nthreads

    cdef int * infections_i
    cdef int * play_infections_i
    cdef int * sub_infections_i
    cdef int * sub_play_infections_i
    cdef int * idxs

    # zero the overal infections
    p = p.start("zero")
    for i in range(0, N_INF_CLASSES):
        infections_i = get_int_array_ptr(infections.work[i])
        play_infections_i = get_int_array_ptr(infections.play[i])

        with nogil, parallel(num_threads=num_threads):
            for j in prange(1, nnodes_plus_one, schedule="static"):
                play_infections_i[j] = 0

            for j in prange(1, nlinks_plus_one, schedule="static"):
                infections_i[j] = 0
    p = p.stop()

    # aggregate from the sub-infections
    for ii, subinf in enumerate(infections.subinfs):
        p = p.start(f"aggregate_{ii}")
        for i in range(0, N_INF_CLASSES):
            infections_i = get_int_array_ptr(infections.work[i])
            play_infections_i = get_int_array_ptr(infections.play[i])
            sub_infections_i = get_int_array_ptr(subinf.work[i])
            sub_play_infections_i = get_int_array_ptr(subinf.play[i])

            # aggregate the work infections
            if subinf.has_different_work_matrix():
                # the subnetwork has a different work matrix
                idxs = get_int_array_ptr(subinf.get_work_index())
                nsublinks_plus_one = subinf.nlinks + 1

                with nogil, parallel(num_threads=num_threads):
                    for j in prange(1, nsublinks_plus_one, schedule="static"):
                        idx = idxs[j]
                        infections_i[idx] = infections_i[idx] + \
                                            sub_infections_i[j]

            else:
                # the subnetwork has the same work matrix
                with nogil, parallel(num_threads=num_threads):
                    for j in prange(1, nlinks_plus_one, schedule="static"):
                        infections_i[j] = infections_i[j] + \
                                          sub_infections_i[j]

            # aggregate the play infections
            with nogil, parallel(num_threads=num_threads):
                for j in prange(1, nnodes_plus_one, schedule="static"):
                    play_infections_i[j] = play_infections_i[j] + \
                                           sub_play_infections_i[j]

        p = p.stop()

    p = p.stop()


def aggregate_networks(network: Networks, profiler: Profiler,
                       nthreads: int = 1) -> None:
    """Aggregate all of the infection data from the demographic
       sub-networks into an overall total set of data
       that is stored in the overall network

       Parameters
       ----------
       network: Networks
         The collection of networks to be aggregated
       profiler: Profiler
         Profiler used to profile the calculation
       nthreads: int
         The number of threads over which to parallelise this work
    """
    from cython.parallel import parallel, prange

    if profiler is None:
        from ._profiler import NullProfiler
        profiler = NullProfiler()

    p = profiler.start("aggregate_network")

    cdef int i = 0
    cdef int j = 0
    cdef int idx = 0
    cdef int num_threads = nthreads

    nodes = network.overall.nodes
    links = network.overall.links

    cdef int nnodes_plus_one = network.overall.nnodes + 1
    cdef int nlinks_plus_one = network.overall.nlinks + 1
    cdef int nsublinks_plus_one = 0

    cdef double * nodes_play_suscept = get_double_array_ptr(nodes.play_suscept)
    cdef double * sub_nodes_play_suscept

    cdef double * links_weight = get_double_array_ptr(links.weight)
    cdef double * links_suscept = get_double_array_ptr(links.suscept)
    cdef double * sub_links_weight
    cdef double * sub_links_suscept
    cdef int * idxs

    # zero the overall data
    p = p.start("zero")
    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nlinks_plus_one):
            links_weight[i] = 0
            links_suscept[i] = 0

        for i in prange(1, nnodes_plus_one):
            nodes_play_suscept[i] = 0
    p = p.stop()

    for ii, subnet in enumerate(network.subnets):
        p = p.start(f"aggregate_work_{ii}")
        sublinks = subnet.links
        sub_links_weight = get_double_array_ptr(sublinks.weight)
        sub_links_suscept = get_double_array_ptr(sublinks.suscept)

        if subnet.has_different_work_matrix():
            # different work matrix, so need to look up indicies
            idxs = get_int_array_ptr(subnet.get_work_index())
            nsublinks_plus_one = subnet.nlinks + 1

            with nogil, parallel(num_threads=num_threads):
                for i in prange(1, nsublinks_plus_one):
                    idx = idxs[i]
                    links_weight[idx] = links_weight[idx] + \
                                        sub_links_weight[i]
                    links_suscept[idx] = links_suscept[idx] + \
                                         sub_links_weight[i]
        else:
            # same work matrix, so simple accumulation
            with nogil, parallel(num_threads=num_threads):
                for i in prange(1, nlinks_plus_one):
                    links_weight[i] = links_weight[i] + sub_links_weight[i]
                    links_suscept[i] = links_suscept[i] + sub_links_suscept[i]

        p = p.stop()

        p = p.start(f"aggregate_play_{ii}")
        subnodes = subnet.nodes
        sub_nodes_play_suscept = get_double_array_ptr(subnodes.play_suscept)

        with nogil, parallel(num_threads=num_threads):
            for i in prange(1, nnodes_plus_one):
                nodes_play_suscept[i] = nodes_play_suscept[i] + \
                                        sub_nodes_play_suscept[i]

        p = p.stop()

    p = p.stop()




"""
.. currentmodule:: metawards.extractors

Functions
=========

.. autosummary::
    :toctree: generated/

    extract_custom
    extract_default
    extract_large
    extract_none
    extract_small

    output_basic
    output_core
    output_core_omp
    output_core_serial
    output_incidence
    output_dispersal
    output_prevalence
    output_trajectory
    output_wards_trajectory

    setup_core
"""

from ._extract_default import *
from ._extract_custom import *
from ._extract_large import *
from ._extract_none import *
from ._extract_small import *

from ._output_basic import *
from ._output_core import *
from ._output_dispersal import *
from ._output_incidence import *
from ._output_prevalence import *
from ._output_trajectory import *
from ._output_wards_trajectory import *




__all__ = ["extract_large"]


def extract_large(**kwargs):
    """This extractor extracts the default files, plus the
       "large" files, e.g. S, E, I and R for each ward.
    """

    from ._output_wards_trajectory import output_wards_trajectory
    from ._extract_default import extract_default

    return extract_default(**kwargs) + [output_wards_trajectory]




from dataclasses import dataclass as _dataclass

from ._variableset import VariableSet
from ._network import Network

__all__ = ["Demographic"]


@_dataclass
class Demographic:
    """This class represents a single demographic"""

    #: The name of this demographic. This will be used as a label,
    #: and should be unique within the Demographics
    name: str = None

    #: The proportion of "work" (fixed movement) members in this
    #: demographic out of the entire population. This can either
    #: be a single number, or a list with a value for every ward.
    #: Whichever way is used, the sum of "work_ratio" for each demographic
    #: must equal 1.0 in each ward (every member of the work population
    #: must be represented)
    work_ratio: float = 0.0

    #: The proportion of "play" (random movement) members in this
    #: demographic out of the entire population. This can also either
    #: be a single number, or a list with a value for every ward.
    #: Whichever way is used, the sum of "play_ratio" for each demographic
    #: must equal 1.0 in each ward (every member of the play population
    #: must be represented)
    play_ratio: float = 0.0

    #: How the parameters for this demographic should be changed compared
    #: to the parameters used for the whole population. This is currently
    #: changed to fixed values, but future developments in VariableSet
    #: will support rules for varying relative to the whole population.
    #: If this is None then this demographic will have the same
    #: parameters as the whole population
    adjustment: VariableSet = None

    def specialise(self, network: Network, profiler=None,
                   nthreads: int = 1):
        """Return a copy of the passed network that has been specialised
           for this demographic. The returned network will
           contain only members of this demographic, with the
           parameters of the network adjusted according to the rules
           of this demographic

           Parameters
           ----------
           network: Network
             The network to be specialised

           Returns
           -------
           network: Network
             The specialised network
        """
        # Start by making a shallow copy of all elements - I really do
        # only want the immediate children of the network to be copied.
        # I will then change what is needed in deep copies - this should
        # save memory for things that don't change
        import copy
        subnet = copy.copy(network)

        # Now create safe copies of the nodes and links. This will
        # shallow copy what it can, and will deep copy variables
        # that will change during a model run
        subnet.nodes = network.nodes.copy()
        subnet.links = network.links.copy()
        subnet.play = network.play.copy()

        # Now we need to adjust the number of susceptibles in each
        # ward according to work_ratio and play_ratio
        subnet.scale_susceptibles(work_ratio=self.work_ratio,
                                  play_ratio=self.play_ratio)

        # Now specialise the parameters for this network
        if self.adjustment is not None:
            subnet.params = network.params.set_variables(self.adjustment)

        subnet.name = self.name

        return subnet



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
from libc.stdint cimport uintptr_t

from .._network import Network
from .._infections import Infections

from ..utils._profiler import Profiler
from ..utils._get_functions import call_function_on_network

from ..utils._ran_binomial cimport _ran_binomial, \
                                   _get_binomial_ptr, binomial_rng

from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["advance_recovery", "advance_recovery_omp",
           "advance_recovery_serial"]


def advance_recovery_omp(network: Network, infections: Infections, rngs,
                         nthreads: int, profiler: Profiler, **kwargs):
    """Advance the model by processing recovery of individual through
       the different stages of the disease (parallel version of the
       function))

       Parameters
       ----------
       network: Network
         The network being modelled
       infections: Infections
         The space that holds all of the  infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    params = network.params

    play_infections = infections.play
    infections = infections.work

    # Copy arguments from Python into C cdef variables
    cdef int N_INF_CLASSES = len(infections)

    cdef int * infections_i
    cdef int * infections_i_plus_one
    cdef int * play_infections_i
    cdef int * play_infections_i_plus_one

    # get the random number generator
    cdef uintptr_t [::1] rngs_view = rngs
    cdef binomial_rng* rng   # pointer to parallel rng

    # create and initialise variables used in the loop
    cdef int num_threads = nthreads
    cdef int thread_id = 0

    cdef double disease_progress = 0.0

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef int i = 0
    cdef int j = 0
    cdef int l = 0
    cdef int inf_ij = 0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all infections in
    ## 'infections' and 'play_infections' and advances
    ## then through the various stages depending on
    ## the result of a random trial

    p = profiler.start("recovery")
    for i in range(N_INF_CLASSES-2, -1, -1):
        # recovery, move through classes backwards (loop down to 0)
        infections_i = get_int_array_ptr(infections[i])
        infections_i_plus_one = get_int_array_ptr(infections[i+1])
        play_infections_i = get_int_array_ptr(play_infections[i])
        play_infections_i_plus_one = get_int_array_ptr(play_infections[i+1])
        disease_progress = params.disease_params.progress[i]

        with nogil, parallel(num_threads=num_threads):
            thread_id = cython.parallel.threadid()
            rng = _get_binomial_ptr(rngs_view[thread_id])

            for j in prange(1, nlinks_plus_one, schedule="static"):
                inf_ij = infections_i[j]

                if inf_ij > 0:
                    l = _ran_binomial(rng, disease_progress, inf_ij)

                    if l > 0:
                        infections_i_plus_one[j] += l
                        infections_i[j] -= l

            for j in prange(1, nnodes_plus_one, schedule="static"):
                inf_ij = play_infections_i[j]

                if inf_ij > 0:
                    l = _ran_binomial(rng, disease_progress, inf_ij)

                    if l > 0:
                        play_infections_i_plus_one[j] += l
                        play_infections_i[j] -= l

        # end of parallel section
    # end of recovery loop
    p = p.stop()


def advance_recovery_serial(network: Network, infections: Infections,
                            rngs, profiler: Profiler, **kwargs):
    """Advance the model by processing recovery of individual through
       the different stages of the disease (serial version of the
       function))

       Parameters
       ----------
       network: Network
         The network being modelled
       infections: Infections
         The space that holds all of the infections
       rngs:
         The list of thread-safe random number generators, one per thread
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """

    params = network.params

    play_infections = infections.play
    infections = infections.work

    # Copy arguments from Python into C cdef variables
    cdef int N_INF_CLASSES = len(infections)

    cdef int * infections_i
    cdef int * infections_i_plus_one
    cdef int * play_infections_i
    cdef int * play_infections_i_plus_one

    # get the random number generator
    cdef binomial_rng* rng = _get_binomial_ptr(rngs[0])

    cdef double disease_progress = 0.0

    cdef int nnodes_plus_one = network.nnodes + 1
    cdef int nlinks_plus_one = network.nlinks + 1

    cdef int i = 0
    cdef int j = 0
    cdef int l = 0
    cdef int inf_ij = 0

    ## Finally(!) we can now declare the actual loop.
    ## This loops in parallel over all infections in
    ## 'infections' and 'play_infections' and advances
    ## then through the various stages depending on
    ## the result of a random trial

    p = profiler.start("recovery")
    for i in range(N_INF_CLASSES-2, -1, -1):
        # recovery, move through classes backwards (loop down to 0)
        infections_i = get_int_array_ptr(infections[i])
        infections_i_plus_one = get_int_array_ptr(infections[i+1])
        play_infections_i = get_int_array_ptr(play_infections[i])
        play_infections_i_plus_one = get_int_array_ptr(play_infections[i+1])
        disease_progress = params.disease_params.progress[i]

        with nogil:
            for j in range(1, nlinks_plus_one):
                inf_ij = infections_i[j]

                if inf_ij > 0:
                    l = _ran_binomial(rng, disease_progress, inf_ij)

                    if l > 0:
                        infections_i_plus_one[j] += l
                        infections_i[j] -= l

            for j in range(1, nnodes_plus_one):
                inf_ij = play_infections_i[j]

                if inf_ij > 0:
                    l = _ran_binomial(rng, disease_progress, inf_ij)

                    if l > 0:
                        play_infections_i_plus_one[j] += l
                        play_infections_i[j] -= l

        # end of parallel section
    # end of recovery loop
    p = p.stop()

def advance_recovery(nthreads: int, **kwargs):
    """Advance the model by processing recovery of individual through
       the different stages of the disease (parallel version of the
       function))

       Parameters
       ----------
       network: Network or Networks
         The network being modelled
       infections: Infections
         The space that holds all of the "work" infections
       rngs:
         The list of thread-safe random number generators, one per thread
       nthreads: int
         The number of threads over which to parallelise the calculation
       profiler: Profiler
         The profiler used to profile this calculation
       kwargs:
         Extra arguments that may be used by other advancers, but which
         are not used by advance_play
    """
    call_function_on_network(nthreads=nthreads,
                             func=advance_recovery_serial,
                             parallel=advance_recovery_omp,
                             **kwargs)




__all__ = ["mix_evenly"]


def mix_evenly(stage: str, **kwargs):
    """This mixer will evenly mix all demographics. This should be
       equivalent to a null model, and produce similar results
       as if the population was all in a single demographic
    """
    if stage == "foi":
        from ._merge_evenly import merge_evenly
        return [merge_evenly]
    else:
        from ._mix_default import mix_default
        return mix_default(stage=stage, **kwargs)




from typing import Union as _Union
from typing import List as _List
from typing import Tuple as _Tuple

from .._network import Network
from .._networks import Networks
from .._population import Population
from .._variableset import VariableSets, VariableSet
from .._outputfiles import OutputFiles

from ._profiler import Profiler
from ._get_functions import MetaFunction

import os as _os

__all__ = ["get_number_of_processes", "run_models"]


def get_number_of_processes(parallel_scheme: str, nprocs: int = None):
    """This function works out how many processes have been set
       by the paralellisation system called 'parallel_scheme'
    """
    if nprocs is None:
        if parallel_scheme == "multiprocessing":
            return 1
        elif parallel_scheme == "mpi4py":
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            nprocs = comm.Get_size()
            return nprocs
        elif parallel_scheme == "scoop":
            raise ValueError(
                f"You must specify the number of processes for "
                f"scoop to parallelise over")
        else:
            raise ValueError(
                f"You must specify the number of processes to "
                f"use for parallel scheme '{parallel_scheme}'")

    if parallel_scheme == "mpi4py":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        n = comm.Get_size()

        if n < nprocs:
            return n
        else:
            return nprocs
    elif parallel_scheme == "scoop":
        return 4
    elif parallel_scheme == "multiprocessing":
        return nprocs
    else:
        raise ValueError(
            f"Unrecognised parallelisation scheme {parallel_scheme}")


def run_models(network: _Union[Network, Networks],
               variables: VariableSets,
               population: Population,
               nprocs: int, nthreads: int, seed: int,
               nsteps: int, output_dir: OutputFiles,
               iterator: MetaFunction = None,
               extractor: MetaFunction = None,
               mixer: MetaFunction = None,
               mover: MetaFunction = None,
               profiler: Profiler = None,
               parallel_scheme: str = "multiprocessing",
               debug_seeds=False) \
        -> _List[_Tuple[VariableSet, Population]]:
    """Run all of the models on the passed Network that are described
       by the passed VariableSets

       Parameters
       ----------
       network: Network or Networks
         The network(s) to model
       variables: VariableSets
         The sets of VariableSet that represent all of the model
         runs to perform
       population: Population
         The initial population for all of the model runs. This also
         contains the starting date and day for the model outbreak
       nprocs: int
         The number of model runs to perform in parallel
       nthreads: int
         The number of threads to parallelise each model run over
       seed: int
         Random number seed which is used to generate random seeds
         for all model runs
       nsteps: int
         The maximum number of steps to perform for each model - this
         will run until the outbreak is over if this is None
       output_dir: OutputFiles
         The OutputFiles that represents the directory in which all
         output should be placed
       iterator: str
         Iterator to load that will be used to iterate the outbreak
       extractor: str
         Extractor to load that will be used to extract information
       mixer: str
         Mixer to load that will be used to mix demographic data
       mover: str
         Mover to load that will be used to move the population between
         different demographics
       profiler: Profiler
         Profiler used to profile the model run
       parallel_scheme: str
         Which parallel scheme (multiprocessing, mpi4py or scoop) to use
         to run multiple model runs in parallel
       debug_seeds: bool (False)
         Set this parameter to force all runs to use the same seed
         (seed) - this is used for debugging and should never be set
         in production runs

       Returns
       -------
       results: List[ tuple(VariableSet, Population)]
         The set of adjustable variables and final population at the
         end of each run
    """

    if len(variables) == 1:
        # no need to do anything complex - just a single run
        params = network.params.set_variables(variables[0])

        network.update(params, profiler=profiler)

        trajectory = network.run(population=population, seed=seed,
                                 nsteps=nsteps,
                                 output_dir=output_dir,
                                 iterator=iterator,
                                 extractor=extractor,
                                 mixer=mixer,
                                 mover=mover,
                                 profiler=profiler,
                                 nthreads=nthreads)

        results = [(variables[0], trajectory)]

        # perform the final summary
        from ._get_functions import get_summary_functions

        if extractor is None:
            from ..extractors._extract_default import extract_default
            extractor = extract_default
        else:
            from ..extractors._extract_custom import build_custom_extractor
            extractor = build_custom_extractor(extractor)

        funcs = get_summary_functions(network=network, results=results,
                                      output_dir=output_dir,
                                      extractor=extractor,
                                      nthreads=nthreads)

        for func in funcs:
            func(network=network, output_dir=output_dir, results=results)

        return results

    # generate the random number seeds for all of the jobs
    # (for testing, we will use the same seed so that I can check
    #  that they are all working)
    seeds = []

    from ._console import Console

    if seed == 0:
        # this is a special mode that a developer can use to force
        # all jobs to use the same random number seed (15324) that
        # is used for comparing outputs. This should NEVER be used
        # for production code
        Console.warning("Using special mode to fix all random number "
                        "seeds to 15324. DO NOT USE IN PRODUCTION!!!")

        for i in range(0, len(variables)):
            seeds.append(15324)

    elif debug_seeds:
        Console.warning(f"Using special model to make all jobs use the "
                        f"Same random number seed {seed}. "
                        f"DO NOT USE IN PRODUCTION!")

        for i in range(0, len(variables)):
            seeds.append(seed)

    else:
        from ._ran_binomial import seed_ran_binomial, ran_int
        rng = seed_ran_binomial(seed)

        # seed the rngs used for the sub-processes using this rng
        for i in range(0, len(variables)):
            seeds.append(ran_int(rng, 10000, 99999999))

    # set the output directories for all of the jobs - this is based
    # on the fingerprint, so should be unique for each job
    outdirs = []

    for v in variables:
        f = v.output_dir()
        d = _os.path.join(output_dir.get_path(), f)

        i = 1
        base = d

        while d in outdirs:
            i += 1
            d = base + "x%03d" % i

        outdirs.append(d)

    outputs = []

    Console.print(
        f"Running **{len(variables)}** jobs using **{nprocs}** process(es)",
        markdown=True)

    if nprocs == 1:
        # no need to use a pool, as we will repeat this calculation
        # several times
        save_network = network.copy()

        for i, variable in enumerate(variables):
            seed = seeds[i]
            outdir = outdirs[i]

            with output_dir.open_subdir(outdir) as subdir:
                Console.print(
                    f"Running parameter set {i+1} of {len(variables)} "
                    f"using seed {seed}")
                Console.print(f"All output written to {subdir.get_path()}")

                with Console.redirect_output(subdir.get_path(),
                                             auto_bzip=output_dir.auto_bzip()):
                    Console.print(f"Running variable set {i+1}")
                    Console.print(f"Variables: {variable}")
                    Console.print(f"Random seed: {seed}")
                    Console.print(f"nthreads: {nthreads}")

                    # no need to do anything complex - just a single run
                    params = network.params.set_variables(variable)

                    network.update(params, profiler=profiler)

                    with Console.spinner("Computing model run") as spinner:
                        try:
                            output = network.run(population=population,
                                                 seed=seed,
                                                 nsteps=nsteps,
                                                 output_dir=subdir,
                                                 iterator=iterator,
                                                 extractor=extractor,
                                                 mixer=mixer,
                                                 mover=mover,
                                                 profiler=profiler,
                                                 nthreads=nthreads)
                            spinner.success()
                        except Exception as e:
                            spinner.failure()
                            error = f"FAILED: {e.__class__} {e}"
                            Console.error(error)
                            output = None

                    if output is not None:
                        outputs.append((variable, output))
                    else:
                        outputs.append((variable, []))

                if output is not None:
                    Console.panel(f"Completed job {i+1} of {len(variables)}\n"
                                  f"{variable}\n"
                                  f"{output[-1]}",
                                  style="alternate")
                else:
                    Console.error(f"Job {i+1} of {len(variables)}\n"
                                  f"{variable}\n"
                                  f"{error}")
            # end of OutputDirs context manager

            if i != len(variables) - 1:
                # still another run to perform, restore the network
                # to the original state
                network = save_network.copy()
        # end of loop over variable sets
    else:
        from ._worker import run_worker

        # create all of the parameters and options to run
        arguments = []

        if isinstance(network, Networks):
            max_nodes = network.overall.nnodes + 1
            max_links = max(network.overall.nlinks, network.overall.nplay) + 1
        else:
            max_nodes = network.nnodes + 1
            max_links = max(network.nlinks, network.nplay) + 1

        try:
            demographics = network.demographics
        except Exception:
            demographics = None

        # give the workers a clean copy of the profiler
        if profiler is None:
            worker_profiler = None
        else:
            worker_profiler = profiler.__class__()

        for i, variable in enumerate(variables):
            seed = seeds[i]
            outdir = outdirs[i]

            arguments.append({
                "params": network.params.set_variables(variable),
                "demographics": demographics,
                "options": {"seed": seed,
                            "output_dir": outdir,
                            "auto_bzip": output_dir.auto_bzip(),
                            "population": population,
                            "nsteps": nsteps,
                            "iterator": iterator,
                            "extractor": extractor,
                            "mixer": mixer,
                            "mover": mover,
                            "profiler": worker_profiler,
                            "nthreads": nthreads,
                            "max_nodes": max_nodes,
                            "max_links": max_links}
            })

        if parallel_scheme == "multiprocessing":
            # run jobs using a multiprocessing pool
            Console.rule("MULTIPROCESSING")
            from multiprocessing import Pool

            results = []

            with Pool(processes=nprocs) as pool:
                for argument in arguments:
                    results.append(pool.apply_async(run_worker, (argument,)))

                for i, result in enumerate(results):
                    with Console.spinner(
                            "Computing model run") as spinner:
                        try:
                            result.wait()
                            output = result.get()
                            spinner.success()
                        except Exception as e:
                            spinner.failure()
                            error = f"FAILED: {e.__class__} {e}"
                            Console.error(error)
                            output = None

                        if output is not None:
                            Console.panel(
                                f"Completed job {i+1} of {len(variables)}\n"
                                f"{variables[i]}\n"
                                f"{output[-1]}",
                                style="alternate")

                            outputs.append((variables[i], output))
                        else:
                            Console.error(f"Job {i+1} of {len(variables)}\n"
                                          f"{variable}\n"
                                          f"{error}")
                            outputs.append((variables[i], []))

        elif parallel_scheme == "mpi4py":
            # run jobs using a mpi4py pool
            Console.rule("MPI")
            Console.print("Running jobs in parallel using a mpi4py pool")
            from mpi4py import futures
            with futures.MPIPoolExecutor(max_workers=nprocs) as pool:
                results = pool.map(run_worker, arguments)

                for i in range(0, len(variables)):
                    with Console.spinner("Computing model run") as spinner:
                        try:
                            output = next(results)
                            spinner.success()
                        except Exception as e:
                            spinner.failure()
                            error = f"FAILED: {e.__class__} {e}"
                            Console.error(error)
                            output = None

                        if output is not None:
                            Console.panel(
                                f"Completed job {i+1} of {len(variables)}\n"
                                f"{variables[i]}\n"
                                f"{output[-1]}",
                                style="alternate")

                            outputs.append((variables[i], output))
                        else:
                            Console.error(f"Job {i+1} of {len(variables)}\n"
                                          f"{variable}\n"
                                          f"{error}")
                            outputs.append((variables[i], []))

        elif parallel_scheme == "scoop":
            # run jobs using a scoop pool
            Console.rule("SCOOP")
            Console.print("Running jobs in parallel using a scoop pool")
            from scoop import futures

            results = []

            for argument in arguments:
                try:
                    results.append(futures.submit(run_worker, argument))
                except Exception as e:
                    Console.error(
                        f"Error submitting calculation: {e.__class__} {e}\n"
                        f"Trying to submit again...")

                    # try again
                    try:
                        results.append(futures.submit(run_worker, argument))
                    except Exception as e:
                        Console.error(
                            f"No - another error: {e.__class__} {e}\n"
                            f"Skipping this job")
                        results.append(None)

            for i in range(0, len(results)):
                with Console.spinner("Computing model run") as spinner:
                    try:
                        output = results[i].result()
                        spinner.success()
                    except Exception as e:
                        spinner.failure()
                        error = f"FAILED: {e.__class__} {e}"
                        Console.error(error)
                        output = None

                    if output is not None:
                        Console.panel(
                            f"Completed job {i+1} of {len(variables)}\n"
                            f"{variables[i]}\n"
                            f"{output[-1]}",
                            style="alternate")

                        outputs.append((variables[i], output))
                    else:
                        Console.error(f"Job {i+1} of {len(variables)}\n"
                                      f"{variable}\n"
                                      f"{error}")
                        outputs.append((variables[i], []))
        else:
            raise ValueError(f"Unrecognised parallelisation scheme "
                             f"{parallel_scheme}.")

    # perform the final summary
    from ._get_functions import get_summary_functions

    if extractor is None:
        from ..extractors._extract_default import extract_default
        extractor = extract_default
    else:
        from ..extractors._extract_custom import build_custom_extractor
        extractor = build_custom_extractor(extractor)

    funcs = get_summary_functions(network=network, results=outputs,
                                  output_dir=output_dir, extractor=extractor,
                                  nthreads=nthreads)

    for func in funcs:
        try:
            func(network=network, output_dir=output_dir,
                 results=outputs, nthreads=nthreads)
        except Exception as e:
            Console.error(f"Error calling {func}: {e.__class__} {e}")

    return outputs



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

from .._network import Network
from .._population import Population
from .._outputfiles import OutputFiles
from .._infections import Infections

from .._workspace import Workspace

from ..utils._get_functions import call_function_on_network
from ..utils._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

from math import sqrt

__all__ = ["output_dispersal", "output_dispersal_serial"]


def output_dispersal_serial(network: Network, population: Population,
                            output_dir: OutputFiles,
                            workspace: Workspace,
                            **kwargs):
    """This will calculate and output the geographic dispersal
       of the outbreak

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra arguments that are ignored by this function
    """

    wards = network.nodes

    # get data from the python objects
    cdef int * total_new_inf_ward = get_int_array_ptr(
                                        workspace.total_new_inf_ward)

    cdef double * wards_x = get_double_array_ptr(wards.x)
    cdef double * wards_y = get_double_array_ptr(wards.y)

    cdef int nnodes_plus_one = network.nnodes + 1

    # variables to accumulate data
    cdef double x = 0.0
    cdef double y = 0.0

    cdef double sum_x = 0.0
    cdef double sum_y = 0.0
    cdef double sum_x2 = 0.0
    cdef double sum_y2 = 0.0

    cdef int total_new = 0
    cdef int newinf = 0

    # now loop over all wards and accumulate the x/y position weighted
    # by new infections
    cdef int i = 0;

    with nogil:
        for i in range(1, nnodes_plus_one):
            newinf = total_new_inf_ward[i]

            if newinf > 0:
                x = wards_x[i]
                y = wards_y[i]

                sum_x += newinf * x
                sum_y += newinf * y

                sum_x2 += newinf * x * x
                sum_y2 += newinf * y * y

                total_new += newinf

    # get the file handles - this will open the files if
    # they have not already been created
    if network.name is not None:
        name = "_" + network.name.replace(" ", "_")
    else:
        name = ""

    mean_xy_file = output_dir.open(f"MeanXY{name}.dat")
    var_xy_file = output_dir.open(f"VarXY{name}.dat")
    dispersal_file = output_dir.open(f"Dispersal{name}.dat")

    timestep = population.day

    if total_new > 0:
        mean_x = sum_x / total_new
        mean_y = sum_y / total_new

        if total_new > 1:
            var_x = (sum_x2 - sum_x*mean_x) / (total_new - 1)
            var_y = (sum_y2 - sum_y*mean_y) / (total_new - 1)
        else:
            var_x = 0.0
            var_y = 0.0

        if var_x + var_y > 0:
            dispersal = sqrt(var_x + var_y)
        elif var_x + var_y > -0.1:
            # numeric rounding error
            dispersal = 0.0
        else:
            from ..utils._console import Console
            Console.warning(f"Negative var? {var_x+var_y}, {var_x} {var_y}")
            dispersal = 0.0
    else:
        mean_x = 0.0
        mean_y = 0.0

        var_x = 0.0
        var_y = 0.0

        dispersal = 0.0

    mean_xy_file.write("%d %f %f\n" % (timestep, mean_x, mean_y))
    var_xy_file.write("%d %f %f\n" % (timestep, var_x, var_y))
    dispersal_file.write("%d %f\n" % (timestep, dispersal))


def output_dispersal(nthreads: int, **kwargs):
    """This will calculate and output the geographic dispersal
       of the outbreak

       Parameters
       ----------
       network: Network
         The network over which the outbreak is being modelled
       population: Population
         The population experiencing the outbreak
       output_dir: OutputFiles
         The directory in which to place all output files
       workspace: Workspace
         A workspace that can be used to extract data
       kwargs
         Extra arguments that are ignored by this function
    """
    call_function_on_network(nthreads=1,
                             func=output_dispersal_serial,
                             call_on_overall=True,
                             **kwargs)



"""
.. currentmodule:: metawards

Classes
=======

.. autosummary::
    :toctree: generated/

    Demographic
    Demographics
    Disease
    Infections
    InputFiles
    Link
    Links
    Network
    Networks
    Node
    Nodes
    OutputFiles
    Parameters
    Population
    Populations
    VariableSet
    VariableSets
    WardInfo
    WardInfos
    Workspace

Functions
=========

.. autosummary::
    :toctree: generated/

    get_version_string

"""

from . import analysis
from . import app
from . import movers
from . import mixers
from . import extractors
from . import iterators
from . import utils
from ._version import get_versions
from ._links import *
from ._nodes import *
from ._wardinfo import *
from ._workspace import *
from ._infections import *
from ._variableset import *
from ._outputfiles import *
from ._networks import *
from ._network import *
from ._population import *
from ._link import *
from ._node import *
from ._inputfiles import *
from ._disease import *
from ._demographics import *
from ._demographic import *
from ._parameters import *
import sys as _sys

if _sys.version_info < (3, 7):
    print("MetaWards requires Python version 3.7 or above.")
    print("Your python is version")
    print(_sys.version)
    _sys.exit(-1)

__all__ = ["get_version_string"]


# import the pyx cython-compiled modules

__manual_version__ = "0.9.0"

_v = get_versions()
__version__ = _v['version']

if __version__.find("untagged") != -1:
    __version__ = __manual_version__

__branch__ = _v['branch']
__repository__ = _v['repository']
__revisionid__ = _v['full-revisionid']
del _v
del get_versions


def _url(url):
    """Simple function to include URLs on OS's that support them in
       console output
    """
    import sys
    if sys.platform == "win32":
        return url
    else:
        return f"[{url}]({url})"


def get_version_string():
    """Return a version string for metawards which can be printed
       into a file or written out to the screen
    """
    from metawards import get_repository
    repository, v = get_repository(error_on_missing=False)

    if repository is None:
        repo_info = f"""
***WARNING: MetaWardsData cannot be found!
Please see {_url('https://metawards.org/model_data')}
for instructions on how to download and install this necessary data.***
"""
    else:
        if v["is_dirty"]:
            dirty = """
***WARNING: This data has not been committed to git.
You may not be able to reproduce this run.***
"""
        else:
            dirty = ""

        repo_info = f"""
### MetaWardsData information
* version: {v['version']}
* repository: {_url(v['repository'])}
* branch: {v['branch']}
{dirty}
"""

    from ._version import get_versions
    v = get_versions()

    if v['version'].find("untagged") != -1:
        # the version couldn't be found - this is likely because this
        # is run as part of github actions or in a disconnected repo.
        # We need to drop back to '__manual_version__'
        v['version'] = __manual_version__

    if v["dirty"]:
        dirty = """
**WARNING: This version has not been committed to git,
so you may not be able to recover the original
source code that was used to generate this run!**
"""
    else:
        dirty = ""

    return f"""
## MetaWards version {v['version']}
## {_url('https://metawards.org')}

### Source information

* repository: {_url(v['repository'])}
* branch: {v['branch']}
* revision: {v['full-revisionid']}
* last modified: {v['date']}
{dirty}
{repo_info}

### Additional information
Visit {_url('https://metawards.org')} for more information
about metawards, its authors and its license
"""


def print_version_string():
    from metawards.utils import Console
    Console.panel(get_version_string(), markdown=True,
                  style="header", width=72, expand=False)


_system_input = input


def input(prompt: str, default="y"):
    """Wrapper for 'input' that returns 'default' if it detected
       that this is being run from within a batch job or other
       service that doesn't have access to a tty
    """
    import sys

    try:
        if sys.stdin.isatty():
            return _system_input(prompt)
        else:
            print(f"Not connected to a console, so having to use the "
                  f"default ({default})")
            return default
    except Exception as e:
        print(f"Unable to get the input: {e.__class__} {e}")
        print(f"Using the default ({default}) instead")
        return default



"""
.. currentmodule:: metawards.movers

Functions
=========

.. autosummary::
    :toctree: generated/

    go_isolate
    go_to

    move_custom
    move_default

"""

from ._go_isolate import *
from ._go_to import *

from ._move_custom import *
from ._move_default import *

# from ._go_on_holiday import *
# from ._go_back_home import *
# from ._go_weekend import *
# from ._go_weekday import *
# from ._go_to_hospital import *
# from ._go_on_lockdown import *
# from ._go_back_to_normal import *



#!/bin/env/python3
#cython: linetrace=False
# MUST ALWAYS DISABLE AS WAY TOO SLOW FOR ITERATE

cimport cython
from cython.parallel import parallel, prange
cimport openmp

from libc.math cimport sqrt, sin, cos, atan2

import math

from .._parameters import Parameters
from .._network import Network

from ._get_array_ptr cimport get_int_array_ptr, get_double_array_ptr

__all__ = ["add_wards_network_distance"]


cdef inline double deg_to_rad(double x) nogil:
    """Convert the passed angle in degrees to radians"""
    # define this here as M_PI is not in every math.h (especially windows)
    M_PI = 3.14159265358979323846
    return x * M_PI / 180


cdef double distance_lat_long(double lon1, double lat1,
                              double lon2, double lat2) nogil:
    """Calculate the distance in kilometers between two lat/lon points on the
       Earth's surface
    """
    cdef double radius = 6378.16  # Earth's radius in km
    cdef double dlon = deg_to_rad(lon2 - lon1)
    cdef double dlat = deg_to_rad(lat2 - lat1)

    # Calculate using the haversin formula
    cdef double sin_dlat_over_2 = sin(dlat / 2.0)
    cdef double sin_dlon_over_2 = sin(dlon / 2.0)

    cdef double a = (sin_dlat_over_2 * sin_dlat_over_2) + \
                    (cos(deg_to_rad(lat1)) * cos(deg_to_rad(lat2)) *
                     sin_dlon_over_2 * sin_dlon_over_2)

    cdef double angle = 2.0 * atan2(sqrt(a), sqrt(1 - a))

    return angle * radius


cdef double distance_x_y(double x1, double y1,
                         double x2, double y2) nogil:
    """Calculate the distance between two x/y points on
       a plane, in the units of the points (hopefully they
       are in kilometers too...)
    """
    cdef double dx = x1 - x2
    cdef double dy = y1 - y2

    dx *= dx
    dy *= dy

    return sqrt(dx + dy)


def add_wards_network_distance(network: Network, nthreads: int = 1):
    """Reads the location data in network.parameters.input_files.position
       and adds those locations to all of the nodes in the passed
       network. Then it calculates all of the distances between
       the nodes for each link and puts that into the
       'distance' property of each link
    """

    # ncov build does not have WEEKEND defined, so not writing this code now
    params = network.params
    wards = network.nodes
    links = network.links
    play = network.play

    line = None

    from ._console import Console

    Console.print("Reading in the positions...")

    cdef int i1 = 0
    cdef double x = 0.0
    cdef double y = 0.0

    cdef double * wards_x = get_double_array_ptr(wards.x)
    cdef double * wards_y = get_double_array_ptr(wards.y)

    cdef scale_for_km = 1.0

    if params.input_files.coordinates == "x/y":
        calc_distance = distance_x_y
        scale_for_km = 0.001
    elif params.input_files.coordinates == "lat/long":
        calc_distance = distance_lat_long
        scale_for_km = 1.0
    else:
        raise ValueError(f"Unrecognised coordinate system "
                         f"{params.input_files.coordinates}")

    try:
        with open(params.input_files.position, "r") as FILE:
            line = FILE.readline()

            while line:
                words = line.split()
                i1 = int(words[0])
                x = float(words[1])
                y = float(words[2])

                wards.x[i1] = x * scale_for_km
                wards.y[i1] = y * scale_for_km

                if (x == 0) and (y == 0):
                    Console.warning(
                        f"WARNING: Position of ward {i1} is 0,0. This "
                        f"doesn't seem right")

                line = FILE.readline()
    except Exception as e:
        raise ValueError(f"{params.input_files.position} is corrupted or "
                         f"unreadable? Error = {e.__class__}: {e}, "
                         f"line = {line}")

    # make sure that all nodes have valid distances
    cdef int i = 0
    cdef int nmissing = 0

    for i in range(1, network.nnodes+1):
        x = wards.x[i]
        y = wards.y[i]

        if (x == 0) and (y == 0):
            nmissing += 1
            if nmissing == 20:
                Console.print(
                      "Not printing any more missing positions as there "
                      "are too many. You should fix the positions file")
            elif nmissing < 20:
                Console.print(
                    f"WARNING: Position of ward {i} does not appear to have "
                    f"been set - position is ({x} {y}).")

    if nmissing > 0:
        Console.print(f"In total the number of wards with missing positions "
                      f"is {nmissing}")

    Console.print("Calculating distances...")

    cdef double total_distance = 0
    cdef double distance, distance2

    cdef int * links_ifrom = get_int_array_ptr(links.ifrom)
    cdef int * links_ito = get_int_array_ptr(links.ito)

    cdef int * play_ifrom = get_int_array_ptr(play.ifrom)
    cdef int * play_ito = get_int_array_ptr(play.ito)

    cdef double * links_distance = get_double_array_ptr(links.distance)
    cdef double * play_distance = get_double_array_ptr(play.distance)

    cdef double x1 = 0.0
    cdef double x2 = 0.0
    cdef double y1 = 0.0
    cdef double y2 = 0.0

    cdef int nlinks_plus_one = network.nlinks + 1
    cdef int nplay_plus_one = network.nplay + 1

    cdef double too_large = 1000000 * scale_for_km

    cdef int ifrom = 0
    cdef int ito = 0

    cdef int num_threads = nthreads

    with nogil, parallel(num_threads=num_threads):
        for i in prange(1, nlinks_plus_one, schedule="static"):
            ifrom = links_ifrom[i]
            ito = links_ito[i]

            x1 = wards_x[ifrom]
            y1 = wards_y[ifrom]

            x2 = wards_x[ito]
            y2 = wards_y[ito]

            if (x1 == 0 and y1 == 0) or (x2 == 0 and y2 == 0):
                links_distance[i] = 0.0
            else:
                # skipping null points
                distance = calc_distance(x1, y1, x2, y2)

                links_distance[i] = distance
                total_distance += distance

                if distance > too_large:
                    with gil:
                        Console.print(
                              f"Large distance between wards {ifrom} "
                              f"and {ito}: {distance} km? {x1},{y1}  "
                              f"{x2},{y2}")

    Console.print(f"Total links distance equals {total_distance}")

    cdef double total_play_distance = 0.0

    if network.nplay > 0:
        with nogil, parallel(num_threads=1):
            for i in prange(1, nplay_plus_one, schedule="static"):
                ifrom = play_ifrom[i]
                ito = play_ito[i]

                x1 = wards_x[ifrom]
                y1 = wards_y[ifrom]

                x2 = wards_x[ito]
                y2 = wards_y[ito]

                if (x1 == 0 and y1 == 0) or (x2 == 0 and y2 == 0):
                    play_distance[i] = 0.0
                else:
                    # skipping null points
                    distance = calc_distance(x1, y1, x2, y2)

                    play_distance[i] = distance
                    total_play_distance += distance

                    if distance > too_large:
                        with gil:
                            Console.print(
                                  f"Large distance between play wards {ifrom} "
                                  f"and {ito}: {distance} km? {x1},{y1}  "
                                  f"{x2},{y2}")

    Console.print(f"Total play distance equals {total_play_distance}")
    Console.print(
            f"Total distance equals {total_distance+total_play_distance}")

    return network




from typing import List as _List
from ..utils._get_functions import MetaFunction

__all__ = ["iterate_default"]


def iterate_default(stage: str, **kwargs) -> _List[MetaFunction]:
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each iteration of the model run.
       This is the default iterator. It models every day as though
       it is a working day.

       Parameters
       ----------
       stage: str
         Which stage of the day is to be modelled

       Returns
       -------
       funcs: List[MetaFunction]
         The list of functions that will be called in sequence
    """

    if stage == "initialise":
        from ._setup_imports import setup_seed_wards
        from ._advance_additional import setup_additional_seeds

        return [setup_seed_wards, setup_additional_seeds]

    elif stage == "setup":
        from ._advance_additional import advance_additional
        return [advance_additional]

    elif stage == "foi":
        from ._advance_foi import advance_foi
        from ._advance_recovery import advance_recovery
        return [advance_foi, advance_recovery]

    elif stage == "infect":
        from ._iterate_weekday import iterate_weekday
        return iterate_weekday(**kwargs)

    else:
        # we don't do anything at the "analyse" or "finalise" stages
        return []




from typing import Union as _Union
from typing import List as _List
from ..utils._get_functions import MetaFunction, accepts_stage

__all__ = ["move_custom", "build_custom_mover"]


def build_custom_mover(custom_function: _Union[str, MetaFunction],
                       parent_name="__main__") -> MetaFunction:
    """Build and return a custom mover from the passed
       function. This will wrap 'extract_mover' around
       the function to double-check that the custom
       function is doing everything correctly

       Parameters
       ----------
       custom_function
         This can either be a function, which will be wrapped and
         returned, or it can be a string. If it is a string then
         we will attempt to locate or import the function associated
         with that string. The search order is;

         1. Is this 'metawards.movers.custom_function'?
         2. Is this 'custom_function' that is already imported'?
         3. Is this a file name in the current path, if yes then
            find the function in that file (either the first function
            called 'extractXXX' or the specified function if
            custom_function is in the form module::function)

        parent_name: str
          This should be the __name__ of the calling function, e.g.
          call this function as build_custom_mover(func, __name__)

        Returns
        -------
        extractor
          The wrapped mover that is suitable for using in the move
          function.
    """
    from ..utils._console import Console

    if isinstance(custom_function, str):
        Console.print(f"Importing a custom mover from {custom_function}")

        # we need to find the function
        import metawards.movers

        # is it metawards.movers.{custom_function}
        try:
            func = getattr(metawards.movers, custom_function)
            return build_custom_mover(func)
        except Exception:
            pass

        # do we have the function in the current namespace?
        import sys
        try:
            func = getattr(sys.modules[__name__], custom_function)
            return build_custom_mover(func)
        except Exception:
            pass

        # how about the __name__ namespace of the caller
        try:
            func = getattr(sys.modules[parent_name], custom_function)
            return build_custom_mover(func)
        except Exception:
            pass

        # how about the __main__ namespace (e.g. if this was loaded
        # in a script)
        try:
            func = getattr(sys.modules["__main__"], custom_function)
            return build_custom_mover(func)
        except Exception:
            pass

        # can we import this function as a file - need to check that
        # the user hasn't written this as module::function
        if custom_function.find("::") != -1:
            parts = custom_function.split("::")
            func_name = parts[-1]
            func_module = "::".join(parts[0:-1])
        else:
            func_name = None
            func_module = custom_function

        from ..utils._import_module import import_module

        module = import_module(func_module)

        if module is None:
            # we cannot find the extractor
            Console.error(
                f"Cannot find the mover '{custom_function}'."
                f"Please make sure this is spelled correctly and "
                f"any python modules/files needed are in the "
                f"PYTHONPATH or current directory")
            raise ImportError(f"Could not import the mover "
                              f"'{custom_function}'")

        if func_name is None:
            # find the last function that starts with 'move'
            import inspect
            funcs = []
            for name, value in inspect.getmembers(module):
                if name.startswith("move"):
                    if hasattr(value, "__call__"):
                        if value.__module__ == module.__name__:
                            # this is a function defined in this module
                            funcs.append(value)

            if len(funcs) > 0:
                func = funcs[0]

                if len(funcs) > 1:
                    Console.warning(
                        f"Multiple possible matching functions: {funcs}. "
                        f"Choosing {func}. Please use the module::function "
                        f"syntax if this is the wrong choice.")
            else:
                func = None

            if func is not None:
                return build_custom_mover(func)

            Console.error(
                f"Could not find any function in the module "
                f"{custom_function} that has a name that starts "
                f"with 'move'. Please manually specify the "
                f"name using the '{custom_function}::your_function syntax")

            raise ImportError(f"Could not import the mover "
                              f"{custom_function}")

        else:
            if hasattr(module, func_name):
                return build_custom_mover(getattr(module, func_name))

            Console.error(
                f"Could not find the function {func_name} in the "
                f"module {func_module}. Check that the spelling "
                f"is correct and that the right version of the module "
                f"is being loaded.")
            raise ImportError(f"Could not import the mover "
                              f"{custom_function}")

    if not hasattr(custom_function, "__call__"):
        Console.error(
            f"Cannot build a mover for {custom_function} "
            f"as it is missing a __call__ function, i.e. it is "
            f"not a function.")
        raise ValueError(f"You can only build custom movers for "
                         f"actual functions... {custom_function}")

    Console.print(f"Building a custom mover for {custom_function}",
                  style="magenta")

    return lambda **kwargs: move_custom(custom_function=custom_function,
                                        **kwargs)


def move_custom(custom_function: MetaFunction,
                stage: str, **kwargs) -> _List[MetaFunction]:
    """This returns the default list of 'go_XXX' functions that
       are called in sequence for each iteration of the model run.
       This provides a custom mover that uses
       'custom_function' passed from the user. This makes
       sure that if 'stage' is not handled by the custom function,
       then the "move_default" functions for that stage
       are correctly called for all stages except "setup"

       Parameters
       ----------
       custom_function: MetaFunction
         A custom user-supplied function that returns the
         functions that the user would like to be called for
         each step.
       stage: str
         The stage of the day/model

       Returns
       -------
       funcs: List[MetaFunction]
         The list of functions that will be called in sequence
    """

    kwargs["stage"] = stage

    if custom_function is None:
        from ._move_default import move_default
        return move_default(**kwargs)

    elif stage == "setup" or accepts_stage(custom_function):
        # most custom functions operate at the 'setup' stage,
        # so movers that don't specify a stage are assumed to
        # only operate here (every other stage is 'move_default')
        return custom_function(**kwargs)

    else:
        from ._move_default import move_default
        return move_default(**kwargs)




from .._network import Network

__all__ = ["fill_in_gaps"]


def fill_in_gaps(network: Network, max_nodes: int):
    """Fills in gaps in the network"""
    nodes = network.nodes
    links = network.links

    cdef int added = 0
    cdef int i = 0
    cdef int link_to = 0
    cdef int [::1] links_ito = links.ito
    cdef int [::1] nodes_label = nodes.label

    cdef int nnodes = network.nnodes
    cdef int MAX_NODES = max_nodes

    from ._console import Console

    for i in range(1, network.nlinks+1):  # careful of 1-indexing
        link_to = links_ito[i]
        if link_to >= MAX_NODES:
            raise MemoryError(f"Adding a link {link_to} that implies we will "
                              f"have more nodes than are preallocated "
                              f"{MAX_NODES}. Increase max_nodes and try "
                              f"again.")

        if nodes_label[link_to] != link_to:
            Console.print(f"ADDING LINK {i} {link_to} {network.nnodes}",
                          style="warning")
            nodes_label[link_to] = link_to

            if link_to > nnodes:
                added += (link_to - nnodes)
                nnodes = link_to


    Console.print(f"Number of added nodes equals {added}", style="info")

    network.nnodes = nnodes




from dataclasses import dataclass as _dataclass
from dataclasses import field as _field
from typing import List as _List
from typing import Dict as _Dict
from typing import Union as _Union
import os as _os
import pathlib as _pathlib

from ._demographic import Demographic
from ._network import Network

__all__ = ["Demographics", "DemographicID", "DemographicIDs"]

_default_demographics_path = _os.path.join(_pathlib.Path.home(),
                                           "GitHub", "MetaWardsData")

_default_folder_name = "demographics"

DemographicID = _Union[str, int]
DemographicIDs = _List[DemographicID]


def _get_value(value):
    """Extract a numeric value from the passed value - this is used
       to allow the demographics.json file to store numbers is
       a variety of formats
    """
    from metawards.utils import safe_eval_number

    if value is None:
        return 0.0

    elif isinstance(value, list):
        lst = []
        for v in value:
            lst.append(safe_eval_number(v))
        return lst

    elif isinstance(value, dict):
        d = []
        for k, v in value.items():
            d[k] = safe_eval_number(v)

        return d
    else:
        return safe_eval_number(value)


@_dataclass(eq=False)
class Demographics:
    """This class holds metadata about all of the demographics
       being modelled
    """
    #: The list of individual Demographic objects, one for each
    #: demographic being modelled
    demographics: _List[Demographic] = _field(default_factory=list)

    #: The random seed to used when using any random number generator
    #: to resolve decisions needed when allocating individuals to
    #: demographics. This is set here so that the Demographics
    #: are uniquely determined and reproducible across runs
    random_seed: int = None

    #: The interaction matrix between demographics. This should
    #: be a list of lists that shows how demographic 'i' affects
    #: demographic 'j'
    interaction_matrix: _List[_List[int]] = None

    #: Map from index to names of demographics - enables lookup by name
    _names: _Dict[str, int] = _field(default_factory=dict)

    _name: str = None
    _version: str = None
    _authors: str = None
    _contacts: str = None
    _references: str = None
    _filename: str = None
    _repository: str = None
    _repository_version: str = None
    _repository_branch: str = None

    def __str__(self):
        d = "\n  ".join([str(x) for x in self.demographics])

        return f"Demographics {self._name}\n" \
               f"loaded from {self._filename}\n" \
               f"version: {self._version}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_branch: {self._repository_branch}\n" \
               f"repository_version: {self._repository_version}\n" \
               f"demographics = [\n  {d}\n]"

    def __len__(self):
        return len(self.demographics)

    def __eq__(self, other):
        if not isinstance(other, Demographics):
            return False

        elif len(self) != len(other):
            return False

        else:
            for name, index in self._names.items():
                if other._names.get(name, None) != index:
                    return False

                if self.demographics[index] != other.demographics[index]:
                    return False

            return True

    def __getitem__(self, item):
        if isinstance(item, str):
            # Lookup by name
            return self.demographics[self._names[item]]
        else:
            # Lookup by index
            return self.demographics[item]

    def copy(self):
        """Return a copy of this demographics object that should
           allow a safe reset between runs. This deepcopies things
           that may change, while shallow copying things that won't
        """
        from copy import copy, deepcopy
        demographics = copy(self)

        demographics.interaction_matrix = deepcopy(self.interaction_matrix)
        demographics.demographics = copy(self.demographics)

        return demographics

    def add(self, demographic: Demographic):
        """Add a demographic to the set to be modelled"""
        if demographic.name is None:
            raise ValueError(
                f"You can only add named demographics to the set.")

        if demographic.name in self._names:
            raise ValueError(
                f"There is already a demographic called "
                f"{demographic.name} in this set. Please rename "
                f"and try again.")

        self.demographics.append(demographic)
        self._names[demographic.name] = len(self.demographics) - 1

    def get_name(self, item):
        """Return the name of the demographic at 'item'"""
        return self.demographics[self.get_index(item)].name

    def get_index(self, item):
        """Return the index of the passed item"""
        try:
            item = int(item)
        except Exception:
            pass

        if isinstance(item, str):
            try:
                return self._names[item]
            except Exception:
                pass

        elif isinstance(item, int):
            try:
                if self.demographics[item] is not None:
                    return item
            except Exception:
                pass

        elif isinstance(item, Demographic):
            for i, d in enumerate(self.demographics):
                if item == d:
                    return i

        # haven't found the item
        raise KeyError(f"There is no demographic is this set that "
                       f"matches {item}.")

    @staticmethod
    def load(name: str = None,
             repository: str = None,
             folder: str = _default_folder_name,
             filename: str = None):
        """Load the parameters for the specified set of demographics.
           This will look for a file called f"{name}.json"
           in the directory f"{repository}/{folder}/{name}.json"

           By default this will load nothing.

           Alternatively you can provide the full path to the
           json file via the "filename" argument

           Parameters
           ----------
           name: str
             The name of the demographics to load. This is the name that
             will be searched for in the METAWARDSDATA diseases directory
           repository: str
             The location of the cloned METAWARDSDATA repository
           folder: str
             The name of the folder within the METAWARDSDATA repository
             that contains the diseases
           filename: str
             The name of the file to load the disease from - this directly
             loads this file without searching through the METAWARDSDATA
             repository

           Returns
           -------
           demographics: Demographics
             The constructed and validated demographics
        """

        repository_version = None
        repository_branch = None

        if filename is None:
            import os
            if os.path.exists(name):
                filename = name
            elif os.path.exists(f"{name}.json"):
                filename = f"{name}.json"

        import os

        if filename is None:
            if repository is None:
                repository = os.getenv("METAWARDSDATA")
                if repository is None:
                    repository = _default_demographics_path

            filename = os.path.join(repository, folder,
                                    f"{name}.json")

            from ._parameters import get_repository_version
            v = get_repository_version(repository)
            repository = v["repository"]
            repository_version = v["version"]
            repository_branch = v["branch"]

        json_file = filename

        try:
            with open(json_file, "r") as FILE:
                import json
                data = json.load(FILE)

        except Exception as e:
            from .utils._console import Console
            Console.error(f"""
Could not find the demographics file {json_file}. "Either it does not exist
or was corrupted. Error was {e.__class__} {e}. To download the disease data
follow the instructions at
[https://metawards.org/model_data](https://metawards.org/model_data).""")

            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        demographics = data.get("demographics", [])
        work_ratios = data.get("work_ratios", [])
        play_ratios = data.get("play_ratios", [])
        random_seed = data.get("random_seed", None)

        if (len(demographics) != len(work_ratios) or
                len(demographics) != len(play_ratios)):
            raise ValueError(
                f"The number of work_ratios ({len(work_ratios)}) must "
                f"equal to number of play_ratios "
                f"({len(play_ratios)}) which must equal the number "
                f"of demographics ({len(demographics)})")

        demos = Demographics(random_seed=random_seed,
                             _name=name,
                             _authors=data.get("author(s)", "unknown"),
                             _contacts=data.get("contact(s)", "unknown"),
                             _references=data.get("reference(s)", "none"),
                             _filename=json_file,
                             _repository=repository,
                             _repository_branch=repository_branch,
                             _repository_version=repository_version)

        for i in range(0, len(demographics)):
            demographic = Demographic(name=demographics[i],
                                      work_ratio=_get_value(work_ratios[i]),
                                      play_ratio=_get_value(play_ratios[i]))
            demos.add(demographic)

        return demos

    def specialise(self, network: Network, profiler=None,
                   nthreads: int = 1):
        """Build the set of networks that will model this set
           of demographics applied to the passed Network.

           Parameters
           ----------
           network: Network
             The overall population model - this contains the base
             parameters, wards, work and play links that define
             the model outbreak
           profiler: Profiler
             Profiler used to profile the specialisation
           nthreads: int
             Number of threads over which to parallelise the work

           Returns
           -------
           networks: Networks
             The set of Networks that represent the model run over the
             full set of different demographics
        """
        if len(self) == 0:
            return network
        else:
            from ._networks import Networks
            return Networks.build(network=network, demographics=self,
                                  profiler=profiler, nthreads=nthreads)



"""
.. currentmodule:: metawards.mixers

Functions
=========

.. autosummary::
    :toctree: generated/

    InteractionMatrix

    merge_evenly

    mix_custom
    mix_default
    mix_evenly
    mix_none

"""

from ._interaction_matrix import *

from ._merge_evenly import *
from ._merge_using_matrix import *

from ._mix_custom import *
from ._mix_default import *
from ._mix_evenly import *
from ._mix_none import *





from ._ran_binomial cimport _construct_binomial_rng, _ran_binomial, \
                            _seed_ran_binomial, _delete_binomial_rng, \
                            _get_binomial_ptr, _ran_uniform


__all__ = ["ran_binomial", "ran_uniform", "ran_int",
           "seed_ran_binomial", "delete_ran_binomial"]


# The plan with this file is to move to an inline cdef that just
# calls the gsl_ran_binomial function directly. Moving away from
# pygsl (and indeed probably gsl) as this is too heavyweight an
# import just to use a random binomial generator. The signature
# of this function in the code will remain though...


def seed_ran_binomial(seed: int = None):
    """Seed and return the random binomial generator. This returns
       the object that you should pass to ran_binomial to generate
       random numbers drawn from a binomial distribution
    """
    if seed is None:
        import random
        seed = random.randint(10000, 99999999)

    from ._console import Console
    Console.print(f"Using random number seed: **{seed}**", markdown=True)
    rng = _construct_binomial_rng()
    _seed_ran_binomial(rng, seed)

    cdef binomial_rng* r = _get_binomial_ptr(rng)

    return rng


def ran_binomial(rng, p: float, n: int):
    """Return a random number drawn from the binomial distribution
       [p,n] (see gsl_ran_binomial for documentation)
    """
    cdef binomial_rng *r = _get_binomial_ptr(rng)
    return _ran_binomial(r, p, n)


def ran_uniform(rng):
    """Return a random double drawn from a uniform distribution between
       zero and one
    """
    cdef binomial_rng *r = _get_binomial_ptr(rng)
    return _ran_uniform(r)


def ran_int(rng, lower=0, upper=(2**32)-1):
    """Draw a random integer from [0,upper] inclusive"""
    return lower + int(ran_uniform(rng) * (upper-lower))


def delete_ran_binomial(rng):
    """Delete the passed random number generator.

       WARNING: It will not be safe to use this generator after
                you have deleted it!
    """
    _delete_binomial_rng(rng)
